# Copyright (c) OpenMMLab. All rights reserved.
import math
from typing import Optional, Sequence, Union

import torch
import torch.nn as nn
import torch.nn.functional as F
from mmcv.cnn import ConvModule
from mmengine.model import BaseModule
from torch import Tensor

from mmflow.registry import MODELS
from mmflow.utils import SampleList, TensorDict, TensorList
from mmflow.utils.typing import OptSampleList
from ..builder import build_components, build_loss
from ..utils import unpack_flow_data_samples
from .base_decoder import BaseDecoder


class CorrelationPyramid(BaseModule):
    """Pyramid Correlation Module.

    The neck of RAFT-Net, which calculates correlation tensor of input features
    with the method of 4D Correlation Pyramid mentioned in RAFT-Net.

    Args:
        num_levels (int): Number of levels in the module.
            Default: 4.
    """

    def __init__(self, num_levels: int = 4) -> None:
        super().__init__()
        self.pool = nn.AvgPool2d(kernel_size=2, stride=2)
        self.num_levels = num_levels

    def forward(self, feat1: Tensor, feat2: Tensor) -> TensorList:
        """Forward function for Correlation pyramid.

        Args:
            feat1 (Tensor): The feature from first input image.
            feat2 (Tensor): The feature from second input image.

        Returns:
            Sequence[Tensor]: The list of correlation which is pooled using
                average pooling with kernel sizes {1, 2, 4, 8}.
        """
        N, C, H, W = feat1.shape
        corr = torch.matmul(
            feat1.view(N, C, -1).permute(0, 2, 1),
            feat2.view(N, C, -1)).view(N, H, W, H, W)
        corr = corr.reshape(N * H * W, 1, H, W) / torch.sqrt(
            torch.tensor(C).float())
        corr_pyramid = [corr]
        for _ in range(self.num_levels - 1):
            _corr = self.pool(corr_pyramid[-1])
            corr_pyramid.append(_corr)

        return corr_pyramid


class MotionEncoder(BaseModule):
    """The module of motion encoder.

    An encoder which consists of several convolution layers and outputs
    features as GRU's input.

    Args:
        num_levels (int): Number of levels used when calculating correlation
            tensor. Default: 4.
        radius (int): Radius used when calculating correlation tensor.
            Default: 4.
        net_type (str): Type of the net. Choices: ['Basic', 'Small'].
            Default: 'Basic'.
    """
    _corr_channels = {'Basic': (256, 192), 'Small': 96}
    _corr_kernel = {'Basic': (1, 3), 'Small': 1}
    _corr_padding = {'Basic': (0, 1), 'Small': 0}

    _flow_channels = {'Basic': (128, 64), 'Small': (64, 32)}
    _flow_kernel = {'Basic': (7, 3), 'Small': (7, 3)}
    _flow_padding = {'Basic': (3, 1), 'Small': (3, 1)}

    _out_channels = {'Basic': 126, 'Small': 80}
    _out_kernel = {'Basic': 3, 'Small': 3}
    _out_padding = {'Basic': 1, 'Small': 1}

    def __init__(self,
                 num_levels: int = 4,
                 radius: int = 4,
                 net_type: str = 'Basic',
                 **kwargs) -> None:
        super().__init__()
        assert net_type in ['Basic', 'Small']
        corr_channels = self._corr_channels.get(net_type) if isinstance(
            self._corr_channels[net_type],
            (tuple, list)) else [self._corr_channels[net_type]]
        corr_kernel = self._corr_kernel.get(net_type) if isinstance(
            self._corr_kernel.get(net_type),
            (tuple, list)) else [self._corr_kernel.get(net_type)]
        corr_padding = self._corr_padding.get(net_type) if isinstance(
            self._corr_padding.get(net_type),
            (tuple, list)) else [self._corr_padding.get(net_type)]

        flow_channels = self._flow_channels.get(net_type)
        flow_kernel = self._flow_kernel.get(net_type)
        flow_padding = self._flow_padding.get(net_type)

        self.out_channels = self._out_channels.get(net_type) if isinstance(
            self._out_channels.get(net_type),
            (tuple, list)) else [self._out_channels.get(net_type)]
        out_kernel = self._out_kernel.get(net_type) if isinstance(
            self._out_kernel.get(net_type),
            (tuple, list)) else [self._out_kernel.get(net_type)]
        out_padding = self._out_padding.get(net_type) if isinstance(
            self._out_padding.get(net_type),
            (tuple, list)) else [self._out_padding.get(net_type)]

        corr_inch = num_levels * (2 * radius + 1)**2
        corr_net = self._make_encoder(corr_inch, corr_channels, corr_kernel,
                                      corr_padding, **kwargs)
        self.corr_net = nn.Sequential(*corr_net)

        flow_inch = 2
        flow_net = self._make_encoder(flow_inch, flow_channels, flow_kernel,
                                      flow_padding, **kwargs)
        self.flow_net = nn.Sequential(*flow_net)

        out_inch = corr_channels[-1] + flow_channels[-1]
        out_net = self._make_encoder(out_inch, self.out_channels, out_kernel,
                                     out_padding, **kwargs)
        self.out_net = nn.Sequential(*out_net)

    def _make_encoder(self, in_channel: int, channels: int, kernels: int,
                      paddings: int, conv_cfg: dict, norm_cfg: dict,
                      act_cfg: dict) -> None:
        """Make encoder."""
        encoder = []

        for ch, k, p in zip(channels, kernels, paddings):

            encoder.append(
                ConvModule(
                    in_channels=in_channel,
                    out_channels=ch,
                    kernel_size=k,
                    padding=p,
                    conv_cfg=conv_cfg,
                    norm_cfg=norm_cfg,
                    act_cfg=act_cfg))
            in_channel = ch
        return encoder

    def forward(self, corr: torch.Tensor, flow: torch.Tensor) -> torch.Tensor:
        """Forward function for MotionEncoder.

        Args:
            corr (Tensor): The correlation feature.
            flow (Tensor): The last estimated optical flow.

        Returns:
            Tensor: The output feature of motion encoder.
        """
        corr_feat = self.corr_net(corr)
        flow_feat = self.flow_net(flow)

        out = self.out_net(torch.cat([corr_feat, flow_feat], dim=1))
        return torch.cat([out, flow], dim=1)


class ConvGRU(BaseModule):
    """GRU with convolution layers.

    GRU cell with fully connected layers replaced with convolutions.

    Args:
        h_channels (int): Number of channels of hidden feature.
        x_channels (int): Number of channels of the concatenation of motion
            feature and context features.
        net_type (str):  Type of the GRU module. Choices: ['Conv', 'SeqConv'].
            Default: 'SeqConv'.
    """
    _kernel = {'Conv': 3, 'SeqConv': ((1, 5), (5, 1))}
    _padding = {'Conv': 1, 'SeqConv': ((0, 2), (2, 0))}

    def __init__(self,
                 h_channels: int,
                 x_channels: int,
                 net_type: str = 'SeqConv') -> None:
        super().__init__()
        assert net_type in ['Conv', 'SeqConv']
        kernel_size = self._kernel.get(net_type) if isinstance(
            self._kernel.get(net_type),
            (tuple, list)) else [self._kernel.get(net_type)]
        padding = self._padding.get(net_type) if isinstance(
            self._padding.get(net_type),
            (tuple, list)) else [self._padding.get(net_type)]

        conv_z = []
        conv_r = []
        conv_q = []

        for k, p in zip(kernel_size, padding):
            conv_z.append(
                ConvModule(
                    in_channels=h_channels + x_channels,
                    out_channels=h_channels,
                    kernel_size=k,
                    padding=p,
                    act_cfg=dict(type='Sigmoid')))
            conv_r.append(
                ConvModule(
                    in_channels=h_channels + x_channels,
                    out_channels=h_channels,
                    kernel_size=k,
                    padding=p,
                    act_cfg=dict(type='Sigmoid')))
            conv_q.append(
                ConvModule(
                    in_channels=h_channels + x_channels,
                    out_channels=h_channels,
                    kernel_size=k,
                    padding=p,
                    act_cfg=dict(type='Tanh')))
        self.conv_z = nn.ModuleList(conv_z)
        self.conv_r = nn.ModuleList(conv_r)
        self.conv_q = nn.ModuleList(conv_q)

    def init_weights(self) -> None:
        """Weights initialization."""

        def weights_init(m):
            classname = m.__class__.__name__
            if classname.find('Conv2d') != -1:
                nn.init.orthogonal_(m.weight)

        self.apply(weights_init)

    def forward(self, h: torch.Tensor, x: torch.Tensor) -> torch.Tensor:
        """Forward function for ConvGRU.

        Args:
            h (Tensor): The last hidden state for GRU block.
            x (Tensor): The current input feature for GRU block

        Returns:
            Tensor: The current hidden state.
        """
        for conv_z, conv_r, conv_q in zip(self.conv_z, self.conv_r,
                                          self.conv_q):

            hx = torch.cat([h, x], dim=1)
            z = conv_z(hx)
            r = conv_r(hx)
            q = conv_q(torch.cat([r * h, x], dim=1))
            h = (1 - z) * h + z * q
        return h


class XHead(BaseModule):
    """A module for flow or mask prediction.

    Args:
        in_channels (int): Input channels of first convolution layer.
        feat_channels (Sequence(int)): List of features channels of different
            convolution layers.
        x_channels (int): Final output channels of predict layer.
        x (str): Type of predict layer. Choice: ['flow', 'mask']
    """

    def __init__(self, in_channels: int, feat_channels: Sequence[int],
                 x_channels: int, x: str) -> None:
        super().__init__()
        conv_layers = []
        for ch in feat_channels:
            conv_layers.append(
                ConvModule(
                    in_channels=in_channels,
                    out_channels=ch,
                    kernel_size=3,
                    padding=1))
            in_channels = ch
        self.layers = nn.Sequential(*conv_layers)
        if x == 'flow':
            self.predict_layer = nn.Conv2d(
                feat_channels[-1], x_channels, kernel_size=3, padding=1)
        elif x == 'mask':
            self.predict_layer = nn.Conv2d(
                feat_channels[-1], x_channels, kernel_size=1, padding=0)
        else:
            raise ValueError(f'x must be \'flow\' or \'mask\', but got {x}')

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Forward Function."""
        x = self.layers(x)
        return self.predict_layer(x)


@MODELS.register_module()
class RAFTDecoder(BaseDecoder):
    """The decoder of RAFT Net.

    The decoder of RAFT Net, which outputs list of upsampled flow estimation.

    Args:
        net_type (str): Type of the net. Choices: ['Basic', 'Small'].
        num_levels (int): Number of levels used when calculating
            correlation tensor.
        radius (int): Radius used when calculating correlation tensor.
        iters (int): Total iteration number of iterative update of RAFTDecoder.
        corr_op_cfg (dict): Config dict of correlation operator.
            Default: dict(type='CorrLookup').
        gru_type (str): Type of the GRU module. Choices: ['Conv', 'SeqConv'].
            Default: 'SeqConv'.
        feat_channels (Sequence(int)): features channels of prediction module.
        mask_channels (int): Output channels of mask prediction layer.
            Default: 64.
        conv_cfg (dict, optional): Config dict of convolution layers in motion
            encoder. Default: None.
        norm_cfg (dict, optional): Config dict of norm layer in motion encoder.
            Default: None.
        act_cfg (dict, optional): Config dict of activation layer in motion
            encoder. Default: None.
    """
    _h_channels = {'Basic': 128, 'Small': 96}
    _cxt_channels = {'Basic': 128, 'Small': 64}

    def __init__(
        self,
        net_type: str,
        num_levels: int,
        radius: int,
        iters: int,
        corr_op_cfg: dict = dict(type='CorrLookup', align_corners=True),
        gru_type: str = 'SeqConv',
        feat_channels: Union[int, Sequence[int]] = 256,
        mask_channels: int = 64,
        conv_cfg: Optional[dict] = None,
        norm_cfg: Optional[dict] = None,
        act_cfg: Optional[dict] = None,
        flow_loss: Optional[dict] = None,
    ) -> None:
        super().__init__()
        assert net_type in ['Basic', 'Small']
        assert type(feat_channels) in (int, tuple, list)
        self.corr_block = CorrelationPyramid(num_levels=num_levels)

        feat_channels = feat_channels if isinstance(tuple,
                                                    list) else [feat_channels]
        self.net_type = net_type
        self.num_levels = num_levels
        self.radius = radius
        self.h_channels = self._h_channels.get(net_type)
        self.cxt_channels = self._cxt_channels.get(net_type)
        self.iters = iters
        self.mask_channels = mask_channels * (2 * radius + 1)
        corr_op_cfg['radius'] = radius
        self.corr_lookup = build_components(corr_op_cfg)
        self.encoder = MotionEncoder(
            num_levels=num_levels,
            radius=radius,
            net_type=net_type,
            conv_cfg=conv_cfg,
            norm_cfg=norm_cfg,
            act_cfg=act_cfg)
        self.gru_type = gru_type
        self.gru = self.make_gru_block()
        self.flow_pred = XHead(self.h_channels, feat_channels, 2, x='flow')

        if net_type == 'Basic':
            self.mask_pred = XHead(
                self.h_channels, feat_channels, self.mask_channels, x='mask')

        if flow_loss is not None:
            self.flow_loss = build_loss(flow_loss)
        self.flow_div = 1.

    def make_gru_block(self):
        """Make GRU block.

        Returns:
            Module: The GRU block.
        """
        return ConvGRU(
            self.h_channels,
            self.encoder.out_channels[0] + 2 + self.cxt_channels,
            net_type=self.gru_type)

    def _upsample(self, flow: Tensor, mask: Optional[Tensor] = None) -> Tensor:
        """Upsample flow field [H/8, W/8, 2] -> [H, W, 2] using convex
        combination.

        Args:
            flow (Tensor): The optical flow with the shape [N, 2, H/8, W/8].
            mask (Tensor, optional): The leanable mask with shape
                [N, grid_size x scale x scale, H/8, H/8].

        Returns:
            Tensor: The output optical flow with the shape [N, 2, H, W].
        """
        scale = 2**(self.num_levels - 1)
        grid_size = self.radius * 2 + 1
        grid_side = int(math.sqrt(grid_size))
        N, _, H, W = flow.shape
        if mask is None:
            new_size = (scale * H, scale * W)
            return scale * F.interpolate(
                flow, size=new_size, mode='bilinear', align_corners=True)
        # predict a (Nx8x8x9xHxW) mask
        mask = mask.view(N, 1, grid_size, scale, scale, H, W)
        mask = torch.softmax(mask, dim=2)

        # extract local grid with 3x3 side  padding = grid_side//2
        upflow = F.unfold(scale * flow, [grid_side, grid_side], padding=1)
        # upflow with shape N, 2, 9, 1, 1, H, W
        upflow = upflow.view(N, 2, grid_size, 1, 1, H, W)

        # take a weighted combination over the neighborhood grid 3x3
        # upflow with shape N, 2, 8, 8, H, W
        upflow = torch.sum(mask * upflow, dim=2)
        upflow = upflow.permute(0, 1, 4, 2, 5, 3)
        return upflow.reshape(N, 2, scale * H, scale * W)

    def forward(self, feat1: Tensor, feat2: Tensor, flow: Tensor,
                h_feat: Tensor, cxt_feat: Tensor) -> TensorList:
        """Forward function for RAFTDecoder.

        Args:
            feat1 (Tensor): The feature from the first input image.
            feat2 (Tensor): The feature from the second input image.
            flow (Tensor): The initialized flow when warm start.
            h_feat (Tensor): The hidden state for GRU cell.
            cxt_feat (Tensor): The contextual feature from the first image.

        Returns:
            Sequence[Tensor]: The list of predicted optical flow.
        """

        corr_pyramid = self.corr_block(feat1, feat2)
        upflow_preds = []
        delta_flow = torch.zeros_like(flow)
        for _ in range(self.iters):
            flow = flow.detach()
            corr = self.corr_lookup(corr_pyramid, flow)
            motion_feat = self.encoder(corr, flow)
            x = torch.cat([cxt_feat, motion_feat], dim=1)
            h_feat = self.gru(h_feat, x)

            delta_flow = self.flow_pred(h_feat)
            flow = flow + delta_flow

            if hasattr(self, 'mask_pred'):
                mask = .25 * self.mask_pred(h_feat)
            else:
                mask = None

            upflow = self._upsample(flow, mask)
            upflow_preds.append(upflow)

        return upflow_preds

    def loss(
        self,
        feat1: Tensor,
        feat2: Tensor,
        flow: Tensor,
        h_feat: Tensor,
        cxt_feat: Tensor,
        data_samples: SampleList,
    ) -> TensorList:
        """Forward function when model training.

        Args:
            feat1 (Tensor): The feature from the first input image.
            feat2 (Tensor): The feature from the second input image.
            flow (Tensor): The last estimated flow from GRU cell.
            h (Tensor): The hidden state for GRU cell.
            cxt_feat (Tensor): The contextual feature from the first image.
            data_samples (list[:obj:`FlowDataSample`]): Each item contains the
                meta information of each image and corresponding annotations.

        Returns:
            Dict[str, Tensor]: The losses of model.
        """

        flow_pred = self.forward(feat1, feat2, flow, h_feat, cxt_feat)

        return self.loss_by_feat(flow_pred, data_samples)

    def predict(self,
                feat1: Tensor,
                feat2: Tensor,
                flow: Tensor,
                h_feat: Tensor,
                cxt_feat: Tensor,
                data_samples: OptSampleList = None) -> SampleList:
        """Forward function when model training.

        Args:
            feat1 (Tensor): The feature from the first input image.
            feat2 (Tensor): The feature from the second input image.
            flow (Tensor): The last estimated flow from GRU cell.
            h (Tensor): The hidden state for GRU cell.
            cxt_feat (Tensor): The contextual feature from the first image.
            data_samples (list[:obj:`FlowDataSample`], optional): Each item
                contains the meta information of each image and corresponding
                annotations. Defaults to None.

        Returns:
            Sequence[FlowDataSample]: The batch of predicted optical flow
                with the same size of images before augmentation.
        """
        flow_pred = self.forward(feat1, feat2, flow, h_feat, cxt_feat)

        flow_result = flow_pred[-1]
        return self.predict_by_feat(flow_result, data_samples)

    def loss_by_feat(self, flow_pred: TensorList,
                     data_samples: SampleList) -> TensorDict:
        """Compute optical flow loss.

        Args:
            flow_pred (Dict[str, Tensor]): multi-level predicted optical flow.
            data_samples (list[:obj:`FlowDataSample`]): Each item contains the
                meta information of each image and corresponding annotations.

        Returns:
            Dict[str, Tensor]: The dict of losses.
        """

        loss = dict()
        batch_gt_flow_fw, _, _, _, batch_gt_valid_fw, _ = \
            unpack_flow_data_samples(data_samples)
        loss['loss_flow'] = self.flow_loss(flow_pred, batch_gt_flow_fw,
                                           batch_gt_valid_fw)
        return loss



# Copyright (c) OpenMMLab. All rights reserved.
from typing import Optional, Sequence, Union

import torch
import torch.nn as nn
from mmcv.cnn import build_conv_layer, build_norm_layer
from mmengine.model import BaseModule

from mmflow.registry import MODELS
from ..utils import BasicBlock, Bottleneck, ResLayer


@MODELS.register_module()
class RAFTEncoder(BaseModule):
    """The feature extraction sub-module in RAFT.

    Args:
        in_channels (int): Number of input channels. Defaults to 3.
        out_channels (int): Number of output channels. Defaults to 128.
        net_type (str): The type of this sub-module, if net_type is Basic, the
            residual block is BasicBlock, if net_type is Small, the residual
            block is Bottleneck. Defaults to 'Basic'.
        stem_channels (int, optional): Number of stem channels. If
            stem_channels is None, it will be set based on net_type. If the
            net_type is Basic, the stem_channels is 64, otherwise the
            stem_channels is 32. Defaults to None.
        base_channels (Sequence[int], optional):  Number of base channels of
            res layer. If base_channels is None, it will be set based on
            net_type. If the net_type is Basic, the base_channels is
            (64, 96, 128), otherwise the base_channels is (8, 16, 24).
            Defaults to None.
        num_stages (int, optional): Resnet stages, if it is None, set
            num_stages as length of base_channels. Defaults to None.
        strides (Sequence[int], optional): Strides of the first block of each
            stage. If it is None, it will be (1, 2, 2). Defaults to None.
        dilations (Sequence[int], optional): Dilation of each stage. If it is
            None, it will be (1, 1, 1). Defaults to None.
        deep_stem (bool): Whether Replace 7x7 conv in input stem with 3 3x3
            conv. Defaults to False.
        avg_down (bool): Whether use AvgPool instead of stride conv when
            downsampling in the bottleneck. Defaults to False.
        frozen_stages (int): Stages to be frozen (stop grad and set eval mode).
            -1 means not freezing any parameters. Defaults to -1.
        conv_cfg (dict, optional): Config dict for convolution layer.
            Default: None..
        norm_cfg (dict, optional): Config dict for each normalization layer.
            Defaults to dict(type='BN', requires_grad=True).
        norm_eval (bool): Whether to set norm layers to eval mode, namely,
            freeze running stats (mean and var). Note: Effect on Batch Norm
            and its variants only. Defaults to False.
        plugins (list[dict], optional): List of plugins for stages, each dict
            contains:

            - cfg (dict, required): Cfg dict to build plugin.

            - position (str, required): Position inside block to insert plugin,
            options: 'after_conv1', 'after_conv2', 'after_conv3'.

            - stages (tuple[bool], optional): Stages to apply plugin, length
            should be same as 'num_stages'
        with_cp (bool): Use checkpoint or not. Using checkpoint will save some
            memory while slowing down the training speed. Defaults to False.
        init_cfg (dict, list, optional): Config of weights initialization.
            Default: None.
    """

    _arch_settings = {
        'Basic': (BasicBlock, (2, 2, 2)),
        'Small': (Bottleneck, (2, 2, 2))
    }

    _stem_channels = {'Basic': 64, 'Small': 32}

    _base_channels = {'Basic': (64, 96, 128), 'Small': (8, 16, 24)}

    _strides = {'Basic': (1, 2, 2), 'Small': (1, 2, 2)}

    _dilations = {'Basic': (1, 1, 1), 'Small': (1, 1, 1)}

    def __init__(self,
                 in_channels: int,
                 out_channels: int,
                 net_type: str = 'Basic',
                 stem_channels: Optional[int] = None,
                 base_channels: Optional[Sequence[int]] = None,
                 num_stages: Optional[int] = None,
                 strides: Optional[Sequence[int]] = None,
                 dilations: Optional[Sequence[int]] = None,
                 deep_stem: bool = False,
                 avg_down: bool = False,
                 frozen_stages: int = -1,
                 conv_cfg: Optional[dict] = None,
                 norm_cfg: dict = dict(type='BN', requires_grad=True),
                 norm_eval: bool = False,
                 plugins: Optional[Sequence[dict]] = None,
                 with_cp: bool = False,
                 init_cfg: Optional[Union[dict, list]] = None) -> None:

        super().__init__(init_cfg=init_cfg)

        if net_type not in self._arch_settings:
            raise KeyError(f'invalid net type {net_type} for RAFT')
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.stem_channels = (
            stem_channels
            if stem_channels is not None else self._stem_channels[net_type])
        self.base_channels = (
            base_channels
            if base_channels is not None else self._base_channels[net_type])
        self.num_stages = (
            num_stages
            if num_stages is not None else len(self._base_channels[net_type]))
        assert self.num_stages >= 1 and self.num_stages <= 3
        self.strides = (
            strides if strides is not None else self._strides[net_type])
        self.dilations = (
            dilations if dilations is not None else self._dilations[net_type])
        assert len(self.strides) == len(self.dilations) == self.num_stages

        self.deep_stem = deep_stem
        self.avg_down = avg_down
        self.frozen_stages = frozen_stages
        self.conv_cfg = conv_cfg
        self.norm_cfg = norm_cfg
        self.norm_eval = norm_eval
        self.plugins = plugins
        self.with_cp = with_cp
        self.block, stage_blocks = self._arch_settings[net_type]
        self.stage_blocks = stage_blocks[:num_stages]
        self.inplanes = self.stem_channels

        self._make_stem_layer(self.in_channels, self.stem_channels)

        self.res_layers = []

        for i, num_blocks in enumerate(self.stage_blocks):
            stride = self.strides[i]
            dilation = self.dilations[i]
            if plugins is not None:
                stage_plugins = self.make_stage_plugins(plugins, i)
            else:
                stage_plugins = None

            planes = self.base_channels[i]
            res_layer = self.make_res_layer(
                block=self.block,
                inplanes=self.inplanes,
                planes=planes,
                num_blocks=num_blocks,
                stride=stride,
                dilation=dilation,
                with_cp=with_cp,
                conv_cfg=conv_cfg,
                norm_cfg=norm_cfg,
                plugins=stage_plugins)
            self.inplanes = planes if net_type == 'Basic' else planes * 4
            layer_name = f'res_layer{i+1}'
            self.add_module(layer_name, res_layer)
            self.res_layers.append(layer_name)

            last_channel = (
                self.base_channels[-1]
                if net_type == 'Basic' else self.base_channels[-1] * 4)
            self.conv2 = build_conv_layer(
                self.conv_cfg, last_channel, out_channels, kernel_size=1)

    def _make_stem_layer(self, in_channels: int, stem_channels: int) -> None:
        """Make stem layer for ResNet."""
        if self.deep_stem:
            self.stem = nn.Sequential(
                build_conv_layer(
                    self.conv_cfg,
                    in_channels,
                    stem_channels // 2,
                    kernel_size=3,
                    stride=2,
                    padding=1,
                    bias=True),
                build_norm_layer(self.norm_cfg, stem_channels // 2)[1],
                nn.ReLU(inplace=True),
                build_conv_layer(
                    self.conv_cfg,
                    stem_channels // 2,
                    stem_channels // 2,
                    kernel_size=3,
                    stride=1,
                    padding=1,
                    bias=True),
                build_norm_layer(self.norm_cfg, stem_channels // 2)[1],
                nn.ReLU(inplace=True),
                build_conv_layer(
                    self.conv_cfg,
                    stem_channels // 2,
                    stem_channels,
                    kernel_size=3,
                    stride=1,
                    padding=1,
                    bias=True),
                build_norm_layer(self.norm_cfg, stem_channels)[1],
                nn.ReLU(inplace=True))
        else:
            self.conv1 = build_conv_layer(
                self.conv_cfg,
                in_channels,
                stem_channels,
                kernel_size=7,
                stride=2,
                padding=3,
                bias=True)
            self.norm1_name, norm1 = build_norm_layer(
                self.norm_cfg, stem_channels, postfix=1)
            self.add_module(self.norm1_name, norm1)
            self.relu = nn.ReLU(inplace=True)

    def make_stage_plugins(self, plugins: Sequence[dict],
                           stage_idx: int) -> Sequence[dict]:
        """make plugins for ResNet 'stage_idx'-th stage .

        Currently we support to insert 'context_block',
        'empirical_attention_block', 'nonlocal_block' into the backbone like
        ResNet/ResNeXt. They could be inserted after conv1/conv2/conv3 of
        Bottleneck.

        An example of plugins format could be :
        >>> plugins=[
        ...     dict(cfg=dict(type='xxx', arg1='xxx'),
        ...          stages=(False, True, True, True),
        ...          position='after_conv2'),
        ...     dict(cfg=dict(type='yyy'),
        ...          stages=(True, True, True, True),
        ...          position='after_conv3'),
        ...     dict(cfg=dict(type='zzz', postfix='1'),
        ...          stages=(True, True, True, True),
        ...          position='after_conv3'),
        ...     dict(cfg=dict(type='zzz', postfix='2'),
        ...          stages=(True, True, True, True),
        ...          position='after_conv3')
        ... ]
        >>> self = ResNet(depth=18)
        >>> stage_plugins = self.make_stage_plugins(plugins, 0)
        >>> assert len(stage_plugins) == 3

        Suppose 'stage_idx=0', the structure of blocks in the stage would be:
            conv1-> conv2->conv3->yyy->zzz1->zzz2
        Suppose 'stage_idx=1', the structure of blocks in the stage would be:
            conv1-> conv2->xxx->conv3->yyy->zzz1->zzz2

        If stages is missing, the plugin would be applied to all stages.

        Args:
            plugins (list[dict]): List of plugins cfg to build. The postfix is
                required if multiple same type plugins are inserted.
            stage_idx (int): Index of stage to build

        Returns:
            list[dict]: Plugins for current stage
        """
        stage_plugins = []
        for plugin in plugins:
            plugin = plugin.copy()
            stages = plugin.pop('stages', None)
            assert stages is None or len(stages) == self.num_stages
            # whether to insert plugin into current stage
            if stages is None or stages[stage_idx]:
                stage_plugins.append(plugin)

        return stage_plugins

    def make_res_layer(self, **kwargs) -> torch.nn.Module:
        """Pack all blocks in a stage into a ``ResLayer``."""
        return ResLayer(**kwargs)

    @property
    def norm1(self) -> torch.nn.Module:
        """nn.Module: the normalization layer named "norm1" """
        return getattr(self, self.norm1_name)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Forward function.

        Args:
            x (torch.Tensor): Input feature.

        Returns:
            torch.Tensor: Output feature.
        """
        if self.deep_stem:
            x = self.stem(x)
        else:
            x = self.conv1(x)
            x = self.norm1(x)
            x = self.relu(x)
        for layer_name in self.res_layers:
            res_layer = getattr(self, layer_name)
            x = res_layer(x)
        out = self.conv2(x)
        return out



# Copyright (c) OpenMMLab. All rights reserved.
from .builder import (build_components, build_decoder, build_encoder,
                      build_flow_estimator, build_operators)
from .data_preprocessor import FlowDataPreprocessor
from .decoders import (FlowNetCDecoder, FlowNetSDecoder, FlowRefine,
                       GMADecoder, IRRPWCDecoder, MaskFlowNetDecoder,
                       MaskFlowNetSDecoder, NetE, OccRefine,
                       OccShuffleUpsample, PWCNetDecoder)
from .encoders import (BasicConvBlock, BasicEncoder, CorrEncoder,
                       FlowNetEncoder, FlowNetSDEncoder, NetC, PWCNetEncoder,
                       RAFTEncoder)
from .flow_estimators import (IRRPWC, RAFT, FlowNet2, FlowNetC, FlowNetCSS,
                              FlowNetS, LiteFlowNet, MaskFlowNet, MaskFlowNetS,
                              PWCNet)
from .losses import (MultiLevelBCE, MultiLevelCharbonnierLoss, MultiLevelEPE,
                     SequenceLoss)
from .utils import (BasicBlock, BasicDenseBlock, BasicLink, Bottleneck,
                    CorrBlock, CorrLookup, DenseLayer, LinkOutput, ResLayer,
                    Warp, occlusion_estimation, unpack_flow_data_samples)

__all__ = [
    'FlowNetEncoder', 'PWCNetEncoder', 'RAFTEncoder', 'NetC',
    'FlowNetSDEncoder', 'CorrEncoder', 'FlowNetCDecoder', 'FlowNetSDecoder',
    'PWCNetDecoder', 'MaskFlowNetSDecoder', 'NetE', 'FlowNetC', 'FlowNetS',
    'LiteFlowNet', 'PWCNet', 'MaskFlowNetS', 'build_encoder', 'build_decoder',
    'build_flow_estimator', 'build_components', 'MultiLevelBCE',
    'MultiLevelEPE', 'MultiLevelCharbonnierLoss', 'SequenceLoss', 'IRRPWC',
    'IRRPWCDecoder', 'FlowRefine', 'OccRefine', 'OccShuffleUpsample',
    'FlowNet2', 'FlowNetCSS', 'MaskFlowNetDecoder', 'MaskFlowNet',
    'GMADecoder', 'build_operators', 'FlowDataPreprocessor', 'RAFT',
    'ResLayer', 'BasicBlock', 'Bottleneck', 'BasicLink', 'LinkOutput',
    'DenseLayer', 'BasicDenseBlock', 'BasicEncoder', 'BasicConvBlock',
    'CorrBlock', 'occlusion_estimation', 'Warp', 'CorrLookup',
    'unpack_flow_data_samples'
]



# training schedule for liteflownet_pre 300k schedule
train_cfg = dict(by_epoch=False, max_iters=800000, val_interval=80000)
val_cfg = dict(type='ValLoop')
test_cfg = dict(type='TestLoop')

# optimizer
optim_wrapper = dict(
    type='OptimWrapper',
    optimizer=dict(
        type='Adam', lr=1.0e-4, weight_decay=4.0e-4, betas=(0.9, 0.999)))

# learning policy
param_scheduler = dict(
    type='MultiStepLR',
    by_epoch=False,
    gamma=0.5,
    milestones=[300000, 500000, 600000, 700000])

default_hooks = dict(
    timer=dict(type='IterTimerHook'),
    logger=dict(type='LoggerHook', interval=50, log_metric_by_epoch=False),
    param_scheduler=dict(type='ParamSchedulerHook'),
    checkpoint=dict(type='CheckpointHook', interval=80000, by_epoch=False),
    sampler_seed=dict(type='DistSamplerSeedHook'),
    visualization=dict(type='FlowVisualizationHook'))



# Copyright (c) OpenMMLab. All rights reserved.
import random
from typing import Any, Dict, Optional, Sequence, Union

import torch
from mmengine.model import BaseDataPreprocessor

from mmflow.registry import MODELS


@MODELS.register_module()
class FlowDataPreprocessor(BaseDataPreprocessor):
    """Image pre-processor for optical flow tasks.

    Comparing with the :class:`mmengine.ImgDataPreprocessor`,

    1. It won't do normalization if ``mean`` is not specified.
    2. It does normalization and color space conversion after stacking batch.
    3. It supports batch augmentations like mixup and cutmix.

    It provides the data pre-processing as follows

    - Collate and move data to the target device.
    - Pad inputs to the maximum size of current batch with defined
      ``pad_value``. The padding size can be divisible by a defined
      ``pad_size_divisor``
    - Stack inputs to batch_inputs.
    - Convert inputs from bgr to rgb if the shape of input is (3, H, W).
    - Normalize image with defined std and mean.
    - Do batch augmentations like Mixup and Cutmix during training.

    Args:
        mean (Sequence[Number], optional): The pixel mean of R, G, B channels.
            Defaults to None.
        std (Sequence[Number], optional): The pixel standard deviation of
            R, G, B channels. Defaults to None.
        bgr_to_rgb (bool): whether to convert image from BGR to RGB.
            Defaults to False.
        rgb_to_bgr (bool): whether to convert image from RGB to RGB.
            Defaults to False.
    """

    def __init__(self,
                 mean: Optional[Sequence[Union[float, int]]] = None,
                 std: Optional[Sequence[Union[float, int]]] = None,
                 sigma_range: Optional[Sequence] = None,
                 clamp_range: Optional[Sequence] = None,
                 bgr_to_rgb: bool = False,
                 rgb_to_bgr: bool = False) -> None:
        super().__init__()
        assert not (bgr_to_rgb and rgb_to_bgr), (
            '`bgr2rgb` and `rgb2bgr` cannot be set to True at the same time')
        assert (mean is None) == (std is None), (
            'mean and std should be both None or tuple')
        self.channel_conversion = rgb_to_bgr or bgr_to_rgb

        if mean is not None:
            assert len(mean) == 3 or len(mean) == 1, (
                '`mean` should have 1 or 3 values, to be compatible with '
                f'RGB or gray image, but got {len(mean)} values')
            assert len(std) == 3 or len(std) == 1, (  # type: ignore
                '`std` should have 1 or 3 values, to be compatible with RGB '  # type: ignore # noqa: E501
                f'or gray image, but got {len(std)} values')
            # Enable the normalization in preprocessing.
            self._enable_normalize = True
            self.register_buffer('mean',
                                 torch.tensor(mean).view(-1, 1, 1), False)
            self.register_buffer('std',
                                 torch.tensor(std).view(-1, 1, 1), False)
        else:
            self._enable_normalize = False

        if sigma_range is not None:
            self.register_buffer('sigma_range', torch.tensor(sigma_range),
                                 False)
            self.register_buffer('clamp_range', torch.tensor(clamp_range),
                                 False)

    def forward(self, data: dict, training: bool = False) -> Dict[str, Any]:
        """Perform normalization、padding and bgr2rgb conversion based on
        ``BaseDataPreprocessor``.

        Args:
            data (dict): data sampled from dataloader.
            training (bool): Whether to enable training time augmentation.

        Returns:
            Dict: Data in the same format as the model input.
        """

        data = self.cast_data(data)  # type: ignore
        inputs = data['inputs']
        data_samples = data.get('data_samples', None)
        img1s = [input_[0, ...] for input_ in inputs]
        img2s = [input_[1, ...] for input_ in inputs]

        if self.channel_conversion and img1s[0].size(0) == 3:
            img1s = [_img1[[2, 1, 0], ...] for _img1 in img1s]
            img2s = [_img2[[2, 1, 0], ...] for _img2 in img2s]

        img1s = [_img1.float() for _img1 in img1s]
        img2s = [_img2.float() for _img2 in img2s]

        if self._enable_normalize:
            img1s = [(img1 - self.mean) / self.std for img1 in img1s]
            img2s = [(img2 - self.mean) / self.std for img2 in img2s]

        if training and hasattr(self, 'sigma_range'):
            # Add Noise
            for i in range(len(img1s)):
                # create new sigma for each image pair
                sigma = torch.tensor(
                    random.uniform(*self.sigma_range), device=self.device)
                img1s[i] = torch.clamp(
                    img1s[i] + torch.randn_like(img1s[i]) * sigma,
                    min=self.clamp_range[0],
                    max=self.clamp_range[1])
                img2s[i] = torch.clamp(
                    img2s[i] + torch.randn_like(img2s[i]) * sigma,
                    min=self.clamp_range[0],
                    max=self.clamp_range[1])

        img1s = torch.stack(img1s, dim=0)
        img2s = torch.stack(img2s, dim=0)

        return dict(
            inputs=torch.cat((img1s, img2s), dim=1), data_samples=data_samples)



dataset_type = 'FlyingChairs'
data_root = 'data/FlyingChairs_release'

global_transform = dict(
    translates=(0.05, 0.05),
    zoom=(1.0, 1.5),
    shear=(0.86, 1.16),
    rotate=(-10., 10.))

relative_transform = dict(
    translates=(0.00375, 0.00375),
    zoom=(0.985, 1.015),
    shear=(1.0, 1.0),
    rotate=(-1.0, 1.0))

# file_client_args = dict(
#     backend='petrel',
#     path_mapping=dict({
#         './data/': 's3://openmmlab/datasets/mmflow/',
#         'data/': 's3://openmmlab/datasets/mmflow/'
#     }))
file_client_args = dict(backend='disk')

train_pipeline = [
    dict(type='LoadImageFromFile', file_client_args=file_client_args),
    dict(type='LoadAnnotations', file_client_args=file_client_args),
    dict(
        type='ColorJitter',
        brightness=0.5,
        contrast=0.5,
        saturation=0.5,
        hue=0.5),
    dict(type='RandomGamma', gamma_range=(0.7, 1.5)),
    dict(type='RandomFlip', prob=0.5, direction='horizontal'),
    dict(type='RandomFlip', prob=0.5, direction='vertical'),
    dict(
        type='RandomAffine',
        global_transform=global_transform,
        relative_transform=relative_transform),
    dict(type='RandomCrop', crop_size=(320, 448)),
    dict(type='PackFlowInputs')
]

test_pipeline = [
    dict(type='LoadImageFromFile', file_client_args=file_client_args),
    dict(type='LoadAnnotations', file_client_args=file_client_args),
    dict(type='InputResize', exponent=6),
    dict(type='PackFlowInputs')
]

flyingchairs_train = dict(
    type=dataset_type,
    pipeline=train_pipeline,
    data_root=data_root,
    split_file='data/FlyingChairs_release/FlyingChairs_train_val.txt')

flyingchairs_test = dict(
    type=dataset_type,
    pipeline=test_pipeline,
    data_root=data_root,
    test_mode=True,
    split_file='data/FlyingChairs_release/FlyingChairs_train_val.txt')

train_dataloader = dict(
    batch_size=1,
    num_workers=2,
    sampler=dict(type='InfiniteSampler', shuffle=True),
    drop_last=True,
    persistent_workers=True,
    dataset=flyingchairs_train)

val_dataloader = dict(
    batch_size=1,
    num_workers=2,
    sampler=dict(type='DefaultSampler', shuffle=False),
    drop_last=False,
    persistent_workers=True,
    dataset=flyingchairs_test)

test_dataloader = val_dataloader
val_evaluator = dict(type='EndPointError')
test_evaluator = val_evaluator



# Copyright (c) OpenMMLab. All rights reserved.
from typing import Dict, Optional, Sequence, Tuple

import torch
import torch.nn as nn
from mmengine.model import BaseModule

from mmflow.registry import MODELS
from mmflow.utils import OptMultiConfig, OptSampleList, SampleList, TensorDict
from ..builder import build_components, build_loss
from ..utils import BasicDenseBlock, CorrBlock, unpack_flow_data_samples
from .base_decoder import BaseDecoder


class PWCModule(BaseModule):
    """Basic module of PWC-Net decoder.

    Args:
        in_channels (int): Input channels of basic dense block.
        up_flow (bool, optional): Whether to calculate upsampling flow and
            feature or not. Default: True.
        densefeat_channels (Sequence[int]): Number of output channels for
            dense layers. Default: (128, 128, 96, 64, 32).
        conv_cfg (dict, optional): Config dict of convolution layer in module.
            Default: None.
        norm_cfg (dict, optional): Config dict of norm layer in module.
            Default: None.
        act_cfg (dict, optional): Config dict of activation layer in module.
            Default: dict(type='LeakyReLU', negative_slope=0.1).
        init_cfg (dict, optional): Config dict of initialization of BaseModule.
            Default: None.
    """

    def __init__(self,
                 in_channels: int,
                 up_flow: bool = True,
                 densefeat_channels: Sequence[int] = (128, 128, 96, 64, 32),
                 conv_cfg: Optional[dict] = None,
                 norm_cfg: Optional[dict] = None,
                 act_cfg: dict = dict(type='LeakyReLU', negative_slope=0.1),
                 init_cfg: OptMultiConfig = None) -> None:
        super().__init__(init_cfg=init_cfg)

        self.up_flow = up_flow
        self.dense_net = BasicDenseBlock(in_channels, densefeat_channels,
                                         conv_cfg, norm_cfg, act_cfg)
        self.last_channels = in_channels + sum(densefeat_channels)
        self.conv_cfg = conv_cfg
        self.norm_cfg = norm_cfg
        self.act_cfg = act_cfg

        self._make_predict_layer()
        self._make_upsample_layer()

    def _make_predict_layer(self) -> torch.nn.Module:
        """Make prediction layer."""
        self.predict_layer = nn.Conv2d(
            self.last_channels, 2, kernel_size=3, padding=1)

    def _make_upsample_layer(self) -> torch.nn.Module:
        """Make upsample  layers."""
        if self.up_flow:
            self.upflow_layer = nn.ConvTranspose2d(
                2, 2, kernel_size=4, stride=2, padding=1)
            self.upfeat_layer = nn.ConvTranspose2d(
                self.last_channels, 2, kernel_size=4, stride=2, padding=1)

    def forward(
        self, x: torch.Tensor
    ) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor]:
        """Forward function for PWCModule.

        Args:
            x (Tensor): The input feature.

        Returns:
            Tuple[Tensor, Tensor, Tensor, Tensor]: The predicted optical flow,
                the feature to predict flow, the upsampled flow from the last
                level, and the upsampled feature.
        """
        feat = self.dense_net(x)
        flow = self.predict_layer(feat)
        upflow = None
        upfeat = None
        if self.up_flow:
            upflow = self.upflow_layer(flow)
            upfeat = self.upfeat_layer(feat)
        return flow, feat, upflow, upfeat


@MODELS.register_module()
class PWCNetDecoder(BaseDecoder):
    """The Decoder of PWC-Net.

    The decoder of PWC-Net which outputs flow predictions and features.

    Args:
        in_channels (dict): Dict of number of input channels for each level.
        densefeat_channels (Sequence[int]): Number of output channels for
            dense layers. Default: (128, 128, 96, 64, 32).
        flow_div (float): The divisor works for scaling the ground truth.
            Default: 20.
        corr_cfg (dict): Config for correlation layer.
            Defaults to dict(type='Correlation', max_displacement=4).
        scaled (bool): Whether to use scaled correlation by the number of
            elements involved to calculate correlation or not.
            Defaults to False.
        warp_cfg (dict): Config for warp operation. Defaults to
            dict(type='Warp', align_corners=True).
        conv_cfg (dict, optional): Config of convolution layer in module.
            Default: None.
        norm_cfg (dict, optional): Config of norm layer in module.
            Default: None.
        act_cfg (dict, optional): Config of activation layer in module.
            Default: dict(type='LeakyReLU', negative_slope=0.1).
        post_processor (dict, optional): Config of flow post process module.
            Default: None
        flow_loss: Config of loss function of optical flow. Default: None.
        init_cfg (dict, list, optional): Config of dict weights initialization.
            Default: None.
    """

    def __init__(self,
                 in_channels: Dict[str, int],
                 densefeat_channels: Sequence[int] = (128, 128, 96, 64, 32),
                 flow_div: float = 20.,
                 corr_cfg: dict = dict(type='Correlation', max_displacement=4),
                 scaled: bool = False,
                 warp_cfg: dict = dict(type='Warp', align_corners=True),
                 conv_cfg: Optional[dict] = None,
                 norm_cfg: Optional[dict] = None,
                 act_cfg: dict = dict(type='LeakyReLU', negative_slope=0.1),
                 post_processor: dict = None,
                 flow_loss: Optional[dict] = None,
                 init_cfg: OptMultiConfig = None) -> None:

        assert isinstance(in_channels, dict)

        super().__init__(init_cfg)

        self.in_channels = in_channels
        self.densefeat_channels = densefeat_channels
        self.flow_div = flow_div

        self.flow_levels = list(in_channels.keys())
        self.flow_levels.sort()
        self.start_level = self.flow_levels[-1]
        self.end_level = self.flow_levels[0]

        self.corr_cfg = corr_cfg
        self.scaled = scaled
        self.warp_cfg = warp_cfg

        self.conv_cfg = conv_cfg
        self.norm_cfg = norm_cfg
        self.act_cfg = act_cfg

        if flow_loss is not None:
            self.flow_loss = build_loss(flow_loss)

        self._make_corr_block(self.corr_cfg, self.act_cfg, self.scaled)

        if warp_cfg is not None:
            self._make_warp(self.warp_cfg)

        self.multiplier = dict()
        for level in self.flow_levels:
            self.multiplier[level] = self.flow_div * 2**(-int(level[-1]))

        self.post_processor = (
            build_components(post_processor)
            if post_processor is not None else None)

        self._make_layers()

    def _make_layers(self) -> None:
        """Build sub-modules of this decoder."""
        layers = []

        for level in self.flow_levels:
            up_flow = (level != self.end_level)
            layers.append([
                level,
                self._make_layer(self.in_channels[level], up_flow=up_flow)
            ])
        self.decoders = nn.ModuleDict(layers)

    def _make_layer(self,
                    in_channels: int,
                    up_flow: bool = True) -> torch.nn.Module:
        """Build module at each level of this decoder.

        Args:
            in_channels (int): The channels of input feature
            up_sample (bool): Whether upsample flow for the next level.
                Defaults to True.

        Returns:
            torch.nn.Module: The sub-module for this decoder.
        """

        return PWCModule(
            in_channels,
            up_flow,
            self.densefeat_channels,
            conv_cfg=self.conv_cfg,
            norm_cfg=self.norm_cfg,
            act_cfg=self.act_cfg)

    def _make_corr_block(self, corr_cfg: dict, act_cfg: dict,
                         scaled: bool) -> None:
        """Make correlation.

        Args:
            corr_cfg (dict): Config for correlation layer.
            act_cfg (dict): Config of activation layer in module.
            scaled (bool): Whether to use scaled correlation by the number of
                elements involved to calculate correlation or not.
        """
        self.corr_block = CorrBlock(
            corr_cfg=corr_cfg, act_cfg=act_cfg, scaled=scaled)

    def _make_warp(self, warp_cfg: dict) -> None:
        """Build warp operator.

        Args:
            warp_cfg (dict): Config for warp operation.
        """
        self.warp = build_components(warp_cfg)

    def forward(self, feat1: TensorDict, feat2: TensorDict) -> TensorDict:
        """Forward function for PWCNetDecoder.

        Args:
            feat1 (Dict[str, Tensor]): The feature pyramid from the first
                image.
            feat2 (Dict[str, Tensor]): The feature pyramid from the second
                image.

        Returns:
            Dict[str, Tensor]: The predicted multi-levels optical flow.
        """

        flow_pred = dict()
        flow = None
        upflow = None
        upfeat = None

        for level in self.flow_levels[::-1]:
            _feat1, _feat2 = feat1[level], feat2[level]

            if level == self.start_level:
                corr_feat = self.corr_block(_feat1, _feat2)
            else:
                warp_feat = self.warp(_feat2, upflow * self.multiplier[level])
                corr_feat_ = self.corr_block(_feat1, warp_feat)
                corr_feat = torch.cat((corr_feat_, _feat1, upflow, upfeat),
                                      dim=1)

            flow, feat, upflow, upfeat = self.decoders[level](corr_feat)

            flow_pred[level] = flow

        if self.post_processor is not None:
            post_flow = self.post_processor(feat)
            flow_pred[self.end_level] = flow_pred[self.end_level] + post_flow

        return flow_pred

    def loss(self, feat1: TensorDict, feat2: TensorDict,
             data_samples: SampleList) -> TensorDict:
        """Forward function when model training.

        Args:
            feat1 (Dict[str, Tensor]): The feature pyramid from the first
                image.
            feat2 (Dict[str, Tensor]): The feature pyramid from the second
                image.
            data_samples (list[:obj:`FlowDataSample`]): Each item contains the
                meta information of each image and corresponding annotations.

        Returns:
            Dict[str, Tensor]: The dict of losses.
        """

        flow_pred = self.forward(feat1, feat2)
        return self.loss_by_feat(flow_pred, data_samples)

    def predict(self,
                feat1: TensorDict,
                feat2: TensorDict,
                data_samples: OptSampleList = None) -> SampleList:
        """Forward function when model testing.

        Args:
            feat1 (Dict[str, Tensor]): The feature pyramid from the first
                image.
            feat2 (Dict[str, Tensor]): The feature pyramid from the second
                image.
            data_samples (list[:obj:`FlowDataSample`], optional): Each item
                contains the meta information of each image and corresponding
                annotations. Defaults to None.
        Returns:
            Sequence[FlowDataSample]: The batch of predicted optical flow
                with the same size of images before augmentation.
        """

        flow_pred = self.forward(feat1, feat2)
        flow_results = flow_pred[self.end_level]
        return self.predict_by_feat(flow_results, data_samples)

    def loss_by_feat(self, flow_pred: TensorDict,
                     data_samples: SampleList) -> TensorDict:
        """Compute optical flow loss.

        Args:
            flow_pred (Dict[str, Tensor]): multi-level predicted optical flow.
            data_samples (list[:obj:`FlowDataSample`]): Each item contains the
                meta information of each image and corresponding annotations.

        Returns:
            Dict[str, Tensor]: The dict of losses.
        """
        loss = dict()
        batch_gt_flow_fw, _, _, _, batch_gt_valid_fw, _ = \
            unpack_flow_data_samples(data_samples)
        loss['loss_flow'] = self.flow_loss(flow_pred, batch_gt_flow_fw,
                                           batch_gt_valid_fw)
        return loss



_base_ = [
    '../_base_/models/liteflownet2/liteflownet2_pre_M5S5R5.py',
    '../_base_/datasets/flyingchairs_320x448.py',
    '../_base_/schedules/liteflownet_pre_240k.py',
    '../_base_/default_runtime.py'
]

# Weights are initialized from model of previous stage
load_from = 'https://download.openmmlab.com/mmflow/liteflownet2/liteflownet2_pre_M6S6R6_8x1_flyingchairs_320x448.pth'  # noqa



train_dataset_type = 'FlyingThings3D'
train_data_root = 'data/flyingthings3d'
test_dataset_type = 'Sintel'
test_data_root = 'data/Sintel'

train_pipeline = [
    dict(type='LoadImageFromFile'),
    dict(type='LoadAnnotations'),
    dict(
        type='ColorJitter',
        asymmetric_prob=0.2,
        brightness=0.4,
        contrast=0.4,
        saturation=0.4,
        hue=0.5 / 3.14),
    dict(type='Erase', prob=0.5, bounds=[50, 100], max_num=3),
    dict(
        type='SpacialTransform',
        spacial_prob=0.8,
        stretch_prob=0.8,
        crop_size=(400, 720),
        min_scale=-0.4,
        max_scale=0.8,
        max_stretch=0.2),
    dict(type='RandomCrop', crop_size=(400, 720)),
    dict(type='RandomFlip', prob=0.5, direction='horizontal'),
    dict(type='RandomFlip', prob=0.1, direction='vertical'),
    dict(type='Validation', max_flow=1000.),
    dict(type='PackFlowInputs')
]
test_pipeline = [
    dict(type='LoadImageFromFile'),
    dict(type='LoadAnnotations'),
    dict(type='InputPad', exponent=3),
    dict(type='PackFlowInputs')
]

train_dataset_cleanpass = dict(
    type=train_dataset_type,
    data_root=train_data_root,
    pipeline=train_pipeline,
    test_mode=False,
    pass_style='clean',
    scene='left',
    double=True)

train_dataset_finalpass = dict(
    type=train_dataset_type,
    data_root=train_data_root,
    pipeline=train_pipeline,
    test_mode=False,
    pass_style='final',
    scene='left',
    double=True)

test_data_cleanpass = dict(
    type=test_dataset_type,
    data_root=test_data_root,
    pipeline=test_pipeline,
    test_mode=False,
    pass_style='clean')

test_data_finalpass = dict(
    type=test_dataset_type,
    data_root=test_data_root,
    pipeline=test_pipeline,
    test_mode=False,
    pass_style='final')

train_dataloader = dict(
    batch_size=2,
    num_workers=5,
    sampler=dict(type='InfiniteSampler', shuffle=True),
    drop_last=True,
    persistent_workers=True,
    dataset=dict(
        type='ConcatDataset',
        datasets=[train_dataset_cleanpass, train_dataset_finalpass]))

val_dataloader = [
    dict(
        batch_size=1,
        num_workers=2,
        sampler=dict(type='DefaultSampler', shuffle=False),
        drop_last=False,
        persistent_workers=True,
        dataset=test_data_cleanpass),
    dict(
        batch_size=1,
        num_workers=2,
        sampler=dict(type='DefaultSampler', shuffle=False),
        drop_last=False,
        persistent_workers=True,
        dataset=test_data_finalpass)
]
test_dataloader = val_dataloader
val_evaluator = [
    dict(type='EndPointError', prefix='clean'),
    dict(type='EndPointError', prefix='final')
]
test_evaluator = val_evaluator



# Copyright (c) OpenMMLab. All rights reserved.
from typing import Optional

import mmcv
import numpy as np
from mmengine.visualization import Visualizer

from mmflow.registry import VISUALIZERS
from mmflow.structures import FlowDataSample


@VISUALIZERS.register_module()
class FlowLocalVisualizer(Visualizer):
    """MMFlow Local Visualizer.

    Args:
        name (str): Name of the instance. Defaults to 'visualizer'.
        image (np.ndarray, optional): the origin image to draw. The format
            should be RGB. Defaults to None.
        vis_backends (list, optional): Visual backend config list.
            Default to None.
        save_dir (str, optional): Save file dir for all storage backends.
            If it is None, the backend storage will not save any data.
        fig_save_cfg (dict): Keyword parameters of figure for saving.
            Defaults to empty dict.
        fig_show_cfg (dict): Keyword parameters of figure for showing.
            Defaults to empty dict.
    """

    def __init__(self, name='visualizer', **kwargs):
        super().__init__(name, **kwargs)

    def add_datasample(self,
                       name: str,
                       image: Optional[np.ndarray] = None,
                       data_sample: Optional[FlowDataSample] = None,
                       draw_gt: bool = True,
                       draw_pred: bool = True,
                       show: bool = False,
                       wait_time: int = 0,
                       step: int = 0) -> None:
        """Draw datasample and save to all backends.

        - If GT and prediction are plotted at the same time, they are
        displayed in a stitched image where the left image is the
        ground truth and the right image is the prediction.
        - If ``show`` is True, all storage backends are ignored, and
        the images will be displayed in a local window.

        Args:
            name (str): The image identifier.
            image (np.ndarray, optional): The image to draw. For MMFlow,
                set to None.
            data_sample (:obj:`FlowDataSample`, optional): The
                annotation data of every samples. Defaults to None.
            draw_gt (bool): Whether to draw GT FlowDataSample. Default to True.
            draw_pred (bool): Whether to draw Prediction FlowDataSample.
                Defaults to True.
            show (bool): Whether to display the drawn image. Default to False.
            wait_time (int): Delay in milliseconds. 0 is the special
                value that means "forever". Defaults to 0.
            step (int): Global step value to record. Defaults to 0.
        """

        gt_flow_fw_map = None
        pred_flow_fw_map = None

        if draw_gt and data_sample is not None and 'gt_flow_fw' in data_sample:
            gt_flow_fw = data_sample.gt_flow_fw.data.permute(1, 2,
                                                             0).cpu().numpy()
            gt_flow_fw_map = np.uint8(mmcv.flow2rgb(gt_flow_fw) * 255.)

        if (draw_pred and data_sample is not None
                and 'pred_flow_fw' in data_sample):
            pred_flow_fw = data_sample.pred_flow_fw.data.permute(
                1, 2, 0).cpu().numpy()
            pred_flow_fw_map = np.uint8(mmcv.flow2rgb(pred_flow_fw) * 255.)

        if gt_flow_fw_map is not None and pred_flow_fw_map is not None:
            assert gt_flow_fw_map.shape == pred_flow_fw_map.shape
            drawn_img = np.concatenate((gt_flow_fw_map, pred_flow_fw_map),
                                       axis=1)
        elif gt_flow_fw_map is not None:
            drawn_img = gt_flow_fw_map
        else:
            drawn_img = pred_flow_fw_map

        if show:
            self.show(drawn_img, win_name=name, wait_time=wait_time)
        else:
            self.add_image(name, drawn_img, step)



# Copyright (c) OpenMMLab. All rights reserved.
from typing import Optional, Tuple

import torch
import torch.nn as nn
from torch import Tensor

from mmflow.registry import MODELS
from mmflow.utils import OptSampleList, SampleList, TensorDict, TensorList
from ..builder import build_encoder
from .pwcnet import PWCNet


@MODELS.register_module()
class RAFT(PWCNet):
    """RAFT model.

    Args:
        num_levels (int): Number of levels in .
        radius (int): Number of radius in  .
        cxt_channels (int): Number of channels of context feature.
        h_channels (int): Number of channels of hidden feature in .
        cxt_encoder (dict): Config dict for building context encoder.
        freeze_bn (bool, optional): Whether to freeze batchnorm layer or not.
            Default: False.
    """

    def __init__(self,
                 num_levels: int,
                 radius: int,
                 cxt_channels: int,
                 h_channels: int,
                 cxt_encoder: dict,
                 freeze_bn: bool = False,
                 **kwargs) -> None:
        super().__init__(**kwargs)
        self.num_levels = num_levels
        self.radius = radius
        self.context = build_encoder(cxt_encoder)
        self.h_channels = h_channels
        self.cxt_channels = cxt_channels

        assert self.num_levels == self.decoder.num_levels
        assert self.radius == self.decoder.radius
        assert self.h_channels == self.decoder.h_channels
        assert self.cxt_channels == self.decoder.cxt_channels
        assert self.h_channels + self.cxt_channels == self.context.out_channels

        if freeze_bn:
            self.freeze_bn()

    def freeze_bn(self) -> None:
        """Set batch normalization layer evaluation mode."""
        for m in self.modules():
            if isinstance(m, nn.BatchNorm2d):
                m.eval()

    def extract_feat(self,
                     imgs: Tensor) -> Tuple[Tensor, Tensor, Tensor, Tensor]:
        """Extract features from images.

        Args:
            imgs (Tensor): The concatenated input images.

        Returns:
            Tuple[Tensor, Tensor, Tensor, Tensor]: The feature from the first
                image, the feature from the second image, the hidden state
                feature for GRU cell and the contextual feature.
        """
        in_channels = self.encoder.in_channels
        img1 = imgs[:, :in_channels, ...]
        img2 = imgs[:, in_channels:, ...]

        feat1 = self.encoder(img1)
        feat2 = self.encoder(img2)
        cxt_feat = self.context(img1)

        h_feat, cxt_feat = torch.split(
            cxt_feat, [self.h_channels, self.cxt_channels], dim=1)
        h_feat = torch.tanh(h_feat)
        cxt_feat = torch.relu(cxt_feat)

        return feat1, feat2, h_feat, cxt_feat

    def loss(
        self,
        inputs: Tensor,
        data_samples: SampleList,
        flow_init: Optional[Tensor] = None,
    ) -> TensorDict:
        """Forward function for RAFT when model training.

        Args:
            inputs (Tensor): The concatenated input images.
            data_samples (list[:obj:`FlowDataSample`]): Each item contains the
                meta information of each image and corresponding annotations.
            flow_init (Tensor, optional): The initialized flow when warm start.
                Default to None.

        Returns:
            Dict[str, Tensor]: The losses of output.
        """

        feat1, feat2, h_feat, cxt_feat = self.extract_feat(inputs)
        B, _, H, W = feat1.shape

        if flow_init is None:
            flow_init = torch.zeros((B, 2, H, W), device=feat1.device)

        return self.decoder.loss(
            feat1,
            feat2,
            flow=flow_init,
            h_feat=h_feat,
            cxt_feat=cxt_feat,
            data_samples=data_samples)

    def _forward(self,
                 inputs: Tensor,
                 data_samples: OptSampleList = None,
                 flow_init=None) -> TensorList:
        """_summary_

        Args:
            inputs (torch.Tensor): The input tensor with shape
                (N, C, ...) in general.
            data_samples (list[:obj:`FlowDataSample`], optional): Each item
                contains the meta information of each image and corresponding
                annotations. Defaults to None.
            flow_init (Tensor, optional): The initialized flow when warm start.
                Default to None.
        Returns:
            TensorList: The list of tensor.
        """
        feat1, feat2, h_feat, cxt_feat = self.extract_feat(inputs)
        B, _, H, W = feat1.shape

        if flow_init is None:
            flow_init = torch.zeros((B, 2, H, W), device=feat1.device)

        return self.decoder(
            feat1, feat2, flow=flow_init, h_feat=h_feat, cxt_feat=cxt_feat)

    def predict(self,
                imgs: Tensor,
                data_samples: OptSampleList = None,
                flow_init: Optional[Tensor] = None) -> SampleList:
        """Forward function for RAFT when model testing.

        Args:
            imgs (Tensor): The concatenated input images.
            data_samples (list[:obj:`FlowDataSample`], optional): Each item
                contains the meta information of each image and corresponding
                annotations. Defaults to None.
            flow_init (Tensor, optional): The initialized flow when warm start.
                Default to None.

        Returns:
            Sequence[FlowDataSample]: The batch of predicted optical flow
                with the same size of images before augmentation.
        """
        train_iter = self.decoder.iters
        if self.test_cfg is not None and self.test_cfg.get(
                'iters') is not None:
            self.decoder.iters = self.test_cfg.get('iters')

        feat1, feat2, h_feat, cxt_feat = self.extract_feat(imgs)
        B, _, H, W = feat1.shape

        if flow_init is None:
            flow_init = torch.zeros((B, 2, H, W), device=feat1.device)

        results = self.decoder.predict(
            feat1=feat1,
            feat2=feat2,
            flow=flow_init,
            h_feat=h_feat,
            cxt_feat=cxt_feat,
            data_samples=data_samples)
        # recover iter in train
        self.decoder.iters = train_iter

        return results



# Copyright (c) OpenMMLab. All rights reserved.
from .hooks import FlowVisualizationHook, LiteFlowNetStageLoadHook
from .loops import MultiTestLoop, MultiValLoop
from .schedulers import (MultiStageLR, MultiStageMomentum,
                         MultiStageParamScheduler)

__all__ = [
    'LiteFlowNetStageLoadHook', 'FlowVisualizationHook', 'MultiTestLoop',
    'MultiValLoop', 'MultiStageLR', 'MultiStageMomentum',
    'MultiStageParamScheduler'
]



# Copyright (c) OpenMMLab. All rights reserved.

from mmengine.structures import BaseDataElement, PixelData


class FlowDataSample(BaseDataElement):
    """A data structure interface of MMFlow. They are used as interfaces
    between different components.

    The attributes in ``FlowDataSample`` are divided into several parts:

        - ``gt_flow_fw``(PixelData): The ground truth of optical flow from img1
            to img2.
        - ``gt_flow_bw``(PixelData): The ground truth of optical flow from img2
            to img1.
        - ``gt_occ_fw``(PixelData): Forward ground truth of occlusion pixel
        map.
        - ``gt_occ_bw``(PixelData): Backward ground truth of occlusion pixel
        map.
        - ``pred_flow_fw``(PixelData): The prediction of optical flow from img1
            to img2.
        - ``pred_flow_bw``(PixelData): The prediction of optical flow from img2
            to img1.
        - ``pred_occ_fw``(PixelData): Forward prediction of occlusion pixel
        map.
        - ``pred_occ_bw``(PixelData): Backward prediction of occlusion pixel
        map.
        - ``gt_valid_fw``(PixelData): The mask for valid pixel maps, which is
        used in dataset or is used for filtered flow by ``Validation`` in data
        transform.
        - ``gt_valid_bw``(PixelData): The mask for valid pixel maps, which is
        used in dataset or is used for filtered flow by ``Validation`` in data
        transform.

    Examples:
        >>> import torch
        >>> import numpy as np
        >>> from mmengine.structures import PixelData
        >>> from mmflow.structures import FlowDataSample

        >>> data_sample = FlowDataSample()
        >>> img_meta = dict(img_shape=(3, 4, 3))
        >>> gt_flow_fw = PixelData(metainfo=img_meta)
        >>> gt_flow_fw.flow = torch.rand((3, 4, 2))
        >>> data_sample.gt_flow_fw = gt_flow_fw
        >>> print(data_sample)
        <FlowDataSample(

            META INFORMATION

            DATA FIELDS
            gt_flow_fw: <PixelData(

                    META INFORMATION
                    img_shape: (3, 4, 3)

                    DATA FIELDS
                    flow: tensor([[[0.8098, 0.4833],
                                [0.3728, 0.2867],
                                [0.2114, 0.5077],
                                [0.9779, 0.3627]],

                                [[0.9561, 0.6763],
                                [0.2590, 0.6944],
                                [0.7470, 0.1223],
                                [0.0703, 0.5286]],

                                [[0.6824, 0.9608],
                                [0.1438, 0.6182],
                                [0.4836, 0.2257],
                                [0.2003, 0.2354]]])
                ) at 0x7efd6cd88150>
            _gt_flow_fw: <PixelData(

                    META INFORMATION
                    img_shape: (3, 4, 3)

                    DATA FIELDS
                    flow: tensor([[[0.8098, 0.4833],
                                [0.3728, 0.2867],
                                [0.2114, 0.5077],
                                [0.9779, 0.3627]],

                                [[0.9561, 0.6763],
                                [0.2590, 0.6944],
                                [0.7470, 0.1223],
                                [0.0703, 0.5286]],

                                [[0.6824, 0.9608],
                                [0.1438, 0.6182],
                                [0.4836, 0.2257],
                                [0.2003, 0.2354]]])
                ) at 0x7efd6cd88150>
        ) at 0x7efd6ca023d0>
    """

    @property
    def gt_flow_fw(self) -> PixelData:
        return self._gt_flow_fw

    @gt_flow_fw.setter
    def gt_flow_fw(self, value: PixelData) -> None:
        self.set_field(value, '_gt_flow_fw', dtype=PixelData)

    @gt_flow_fw.deleter
    def gt_flow_fw(self) -> None:
        del self._gt_flow_fw

    @property
    def gt_flow_bw(self) -> PixelData:
        return self._gt_flow_bw

    @gt_flow_bw.setter
    def gt_flow_bw(self, value: PixelData) -> None:
        self.set_field(value, '_gt_flow_bw', dtype=PixelData)

    @gt_flow_bw.deleter
    def gt_flow_bw(self) -> None:
        del self._gt_flow_bw

    @property
    def gt_occ_fw(self) -> PixelData:
        return self._gt_occ_fw

    @gt_occ_fw.setter
    def gt_occ_fw(self, value: PixelData) -> None:
        self.set_field(value, '_gt_occ_fw', dtype=PixelData)

    @gt_occ_fw.deleter
    def gt_occ_fw(self) -> None:
        del self._gt_occ_fw

    @property
    def gt_occ_bw(self) -> PixelData:
        return self._gt_occ_bw

    @gt_occ_bw.setter
    def gt_occ_bw(self, value: PixelData) -> None:
        self.set_field(value, '_gt_occ_bw', dtype=PixelData)

    @gt_occ_bw.deleter
    def gt_occ_bw(self) -> None:
        del self._gt_occ_bw

    @property
    def pred_flow_fw(self) -> PixelData:
        return self._pred_flow_fw

    @pred_flow_fw.setter
    def pred_flow_fw(self, value: PixelData) -> None:
        self.set_field(value, '_pred_flow_fw', dtype=PixelData)

    @pred_flow_fw.deleter
    def pred_flow_fw(self) -> None:
        del self._pred_flow_fw

    @property
    def pred_flow_bw(self) -> PixelData:
        return self._pred_flow_bw

    @pred_flow_bw.setter
    def pred_flow_bw(self, value: PixelData) -> None:
        self.set_field(value, '_pred_flow_bw', dtype=PixelData)

    @pred_flow_bw.deleter
    def pred_flow_bw(self) -> None:
        del self._pred_flow_bw

    @property
    def pred_occ_fw(self) -> PixelData:
        return self._pred_occ_fw

    @pred_occ_fw.setter
    def pred_occ_fw(self, value: PixelData) -> None:
        self.set_field(value, '_pred_occ_fw', dtype=PixelData)

    @pred_occ_fw.deleter
    def pred_occ_fw(self) -> None:
        del self._pred_occ_fw

    @property
    def pred_occ_bw(self) -> PixelData:
        return self._pred_occ_bw

    @pred_occ_bw.setter
    def pred_occ_bw(self, value: PixelData) -> None:
        self.set_field(value, '_pred_occ_bw', dtype=PixelData)

    @pred_occ_bw.deleter
    def pred_occ_bw(self):
        del self._pred_occ_bw

    @property
    def gt_valid_fw(self) -> PixelData:
        return self._gt_valid_fw

    @gt_valid_fw.setter
    def gt_valid_fw(self, value: PixelData) -> None:
        self.set_field(value, '_gt_valid_fw', dtype=PixelData)

    @gt_valid_fw.deleter
    def gt_valid_fw(self) -> None:
        del self._gt_valid_fw

    @property
    def gt_valid_bw(self) -> PixelData:
        return self._gt_valid_bw

    @gt_valid_bw.setter
    def gt_valid_bw(self, value: PixelData) -> None:
        self.set_field(value, '_gt_valid_bw', dtype=PixelData)

    @gt_valid_bw.deleter
    def gt_valid_bw(self) -> None:
        del self._gt_valid_bw



_base_ = [
    '../_base_/models/gma/gma_plus-p.py',
    '../_base_/datasets/kitti2015_raft_288x960.py',
    '../_base_/schedules/raft_50k.py', '../_base_/default_runtime.py'
]

model = dict(
    decoder=dict(
        type='GMADecoder',
        net_type='Basic',
        num_levels=4,
        radius=4,
        iters=12,
        corr_op_cfg=dict(type='CorrLookup', align_corners=True),
        gru_type='SeqConv',
        heads=1,
        motion_channels=128,
        position_only=False,
        max_pos_size=160,
        flow_loss=dict(type='SequenceLoss', gamma=0.85),
        act_cfg=dict(type='ReLU')),
    freeze_bn=False,
    test_cfg=dict(iters=32))

# Load model training on mixed datasets and finetune it on KITTI2015s
load_from = 'https://download.openmmlab.com/mmflow/gma/gma_plus-p_8x2_120k_mixed_368x768.pth'  # noqa



# Copyright (c) OpenMMLab. All rights reserved.

import re
from io import BytesIO
from typing import Tuple

import cv2
import matplotlib.pyplot as plt
import mmcv
import numpy as np
from numpy import ndarray


def read_flow(name: str) -> np.ndarray:
    """Read flow file with the suffix '.flo'.

    This function is modified from
    https://lmb.informatik.uni-freiburg.de/resources/datasets/IO.py
    Copyright (c) 2011, LMB, University of Freiburg.

    Args:
        name (str): Optical flow file path.

    Returns:
        ndarray: Optical flow
    """

    with open(name, 'rb') as f:

        header = f.read(4)
        if header.decode('utf-8') != 'PIEH':
            raise Exception('Flow file header does not contain PIEH')

        width = np.fromfile(f, np.int32, 1).squeeze()
        height = np.fromfile(f, np.int32, 1).squeeze()

        flow = np.fromfile(f, np.float32, width * height * 2).reshape(
            (height, width, 2))

    return flow


def write_flow(flow: np.ndarray, flow_file: str) -> None:
    """Write the flow in disk.

    This function is modified from
    https://lmb.informatik.uni-freiburg.de/resources/datasets/IO.py
    Copyright (c) 2011, LMB, University of Freiburg.

    Args:
        flow (ndarray): The optical flow that will be saved.
        flow_file (str): The file for saving optical flow.
    """

    with open(flow_file, 'wb') as f:
        f.write('PIEH'.encode('utf-8'))
        np.array([flow.shape[1], flow.shape[0]], dtype=np.int32).tofile(f)
        flow = flow.astype(np.float32)
        flow.tofile(f)


def visualize_flow(flow: np.ndarray, save_file: str = None) -> np.ndarray:
    """Flow visualization function.

    Args:
        flow (ndarray): The flow will be render
        save_dir ([type], optional): save dir. Defaults to None.
    Returns:
        ndarray: flow map image with RGB order.
    """

    # return value from mmcv.flow2rgb is [0, 1.] with type np.float32
    flow_map = np.uint8(mmcv.flow2rgb(flow) * 255.)
    if save_file:
        plt.imsave(save_file, flow_map)
    return flow_map


def render_color_wheel(save_file: str = 'color_wheel.png') -> np.ndarray:
    """Render color wheel.

    Args:
        save_file (str): The saved file name . Defaults to 'color_wheel.png'.

    Returns:
        ndarray: color wheel image.
    """
    x0 = 75
    y0 = 75
    height = 151
    width = 151
    flow = np.zeros((height, width, 2), dtype=np.float32)

    grid_x = np.tile(np.expand_dims(np.arange(width), 0), [height, 1])
    grid_y = np.tile(np.expand_dims(np.arange(height), 1), [1, width])

    grid_x0 = np.tile(np.array([x0]), [height, width])
    grid_y0 = np.tile(np.array([y0]), [height, width])

    flow[:, :, 0] = grid_x - grid_x0
    flow[:, :, 1] = grid_y - grid_y0

    return visualize_flow(flow, save_file)


def read_flow_kitti(name: str) -> Tuple[np.ndarray, np.ndarray]:
    """Read sparse flow file from KITTI dataset.

    This function is modified from
    https://github.com/princeton-vl/RAFT/blob/master/core/utils/frame_utils.py.
    Copyright (c) 2020, princeton-vl
    Licensed under the BSD 3-Clause License

    Args:
        name (str): The flow file

    Returns:
        Tuple[ndarray, ndarray]: flow and valid map
    """
    # to specify not to change the image depth (16bit)
    flow = cv2.imread(name, cv2.IMREAD_ANYDEPTH | cv2.IMREAD_COLOR)
    flow = flow[:, :, ::-1].astype(np.float32)
    # flow shape (H, W, 2) valid shape (H, W)
    flow, valid = flow[:, :, :2], flow[:, :, 2]
    flow = (flow - 2**15) / 64.0
    return flow, valid


def write_flow_kitti(uv: np.ndarray, filename: str):
    """Write the flow in disk.

    This function is modified from
    https://github.com/princeton-vl/RAFT/blob/master/core/utils/frame_utils.py.
    Copyright (c) 2020, princeton-vl
    Licensed under the BSD 3-Clause License

    Args:
        uv (ndarray): The optical flow that will be saved.
        filename ([type]): The file for saving optical flow.
    """
    uv = 64.0 * uv + 2**15
    valid = np.ones([uv.shape[0], uv.shape[1], 1])
    uv = np.concatenate([uv, valid], axis=-1).astype(np.uint16)
    cv2.imwrite(filename, uv[..., ::-1])


def flow_from_bytes(content: bytes, suffix: str = 'flo') -> ndarray:
    """Read dense optical flow from bytes.

    .. note::
        This load optical flow function works for FlyingChairs, FlyingThings3D,
        Sintel, FlyingChairsOcc datasets, but cannot load the data from
        ChairsSDHom.

    Args:
        content (bytes): Optical flow bytes got from files or other streams.

    Returns:
        ndarray: Loaded optical flow with the shape (H, W, 2).
    """

    assert suffix in ('flo', 'pfm'), 'suffix of flow file must be `flo` '\
        f'or `pfm`, but got {suffix}'

    if suffix == 'flo':
        return flo_from_bytes(content)
    else:
        return pfm_from_bytes(content)


def flo_from_bytes(content: bytes):
    """Decode bytes based on flo file.

    Args:
        content (bytes): Optical flow bytes got from files or other streams.

    Returns:
        ndarray: Loaded optical flow with the shape (H, W, 2).
    """

    # header in first 4 bytes
    header = content[:4]
    if header != b'PIEH':
        raise Exception('Flow file header does not contain PIEH')
    # width in second 4 bytes
    width = np.frombuffer(content[4:], np.int32, 1).squeeze()
    # height in third 4 bytes
    height = np.frombuffer(content[8:], np.int32, 1).squeeze()
    # after first 12 bytes, all bytes are flow
    flow = np.frombuffer(content[12:], np.float32, width * height * 2).reshape(
        (height, width, 2))

    return flow


def pfm_from_bytes(content: bytes) -> np.ndarray:
    """Load the file with the suffix '.pfm'.

    Args:
        content (bytes): Optical flow bytes got from files or other streams.

    Returns:
        ndarray: The loaded data
    """

    file = BytesIO(content)

    color = None
    width = None
    height = None
    scale = None
    endian = None

    header = file.readline().rstrip()
    if header == b'PF':
        color = True
    elif header == b'Pf':
        color = False
    else:
        raise Exception('Not a PFM file.')

    dim_match = re.match(rb'^(\d+)\s(\d+)\s$', file.readline())
    if dim_match:
        width, height = list(map(int, dim_match.groups()))
    else:
        raise Exception('Malformed PFM header.')

    scale = float(file.readline().rstrip())
    if scale < 0:  # little-endian
        endian = '<'
        scale = -scale
    else:
        endian = '>'  # big-endian

    data = np.frombuffer(file.read(), endian + 'f')
    shape = (height, width, 3) if color else (height, width)

    data = np.reshape(data, shape)
    data = np.flipud(data)
    return data[:, :, :-1]


def read_pfm(file: str) -> np.ndarray:
    """Load the file with the suffix '.pfm'.

    This function is modified from
    https://lmb.informatik.uni-freiburg.de/resources/datasets/IO.py
    Copyright (c) 2011, LMB, University of Freiburg.

    Args:
        file (str): The file name will be loaded

    Returns:
        ndarray: The loaded data
    """
    file = open(file, 'rb')

    color = None
    width = None
    height = None
    scale = None
    endian = None

    header = file.readline().rstrip()
    if header.decode('ascii') == 'PF':
        color = True
    elif header.decode('ascii') == 'Pf':
        color = False
    else:
        raise Exception('Not a PFM file.')

    dim_match = re.match(r'^(\d+)\s(\d+)\s$', file.readline().decode('ascii'))
    if dim_match:
        width, height = list(map(int, dim_match.groups()))
    else:
        raise Exception('Malformed PFM header.')

    scale = float(file.readline().decode('ascii').rstrip())
    if scale < 0:  # little-endian
        endian = '<'
        scale = -scale
    else:
        endian = '>'  # big-endian

    data = np.fromfile(file, endian + 'f')
    shape = (height, width, 3) if color else (height, width)

    data = np.reshape(data, shape)
    data = np.flipud(data)
    return data[:, :, :-1]



default_scope = 'mmflow'

env_cfg = dict(
    cudnn_benchmark=False,
    mp_cfg=dict(mp_start_method='fork', opencv_num_threads=0),
    dist_cfg=dict(backend='nccl'),
)

vis_backends = [dict(type='LocalVisBackend')]
visualizer = dict(
    type='FlowLocalVisualizer', vis_backends=vis_backends, name='visualizer')

log_processor = dict(by_epoch=False)
log_level = 'INFO'
resume = False



# Copyright (c) OpenMMLab. All rights reserved.
from collections import OrderedDict
from typing import Tuple

import torch
import torch.distributed as dist
from torch import Tensor

from mmflow.registry import MODELS
from mmflow.utils import TensorDict
from .pwcnet import PWCNet


@MODELS.register_module()
class IRRPWC(PWCNet):
    """IRR-PWC model."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def extract_feat(self, imgs: Tensor) -> Tuple[TensorDict, TensorDict]:
        """Extract features from images.

        Args:
            imgs (Tensor): The concatenated input images.

        Returns:
            Tuple[Dict[str, Tensor], Dict[str, Tensor]]: The feature pyramid of
                the first input image and the feature pyramid of seconds input
                image.
        """
        in_channels = self.encoder.in_channels
        img1 = imgs[:, :in_channels, ...]
        img2 = imgs[:, in_channels:, ...]
        feat1 = self.encoder(img1)
        feat2 = self.encoder(img2)
        feat1['level0'] = img1
        feat2['level0'] = img2
        return feat1, feat2

    @staticmethod
    def _parse_losses(losses):
        """Parse the raw outputs (losses) of the IRRPWC network.

        Args:
            losses (dict): Raw output of the network, which usually contain
                losses and other necessary information.

        Returns:
            tuple[Tensor, dict]: (loss, log_vars), loss is the loss tensor
                which may be a weighted sum of all losses, log_vars contains
                all the variables to be sent to the logger.
        """

        log_vars = OrderedDict()
        for loss_name, loss_value in losses.items():
            if isinstance(loss_value, torch.Tensor):
                log_vars[loss_name] = loss_value.mean()
            elif isinstance(loss_value, list):
                log_vars[loss_name] = sum(_loss.mean() for _loss in loss_value)
            else:
                raise TypeError(
                    f'{loss_name} is not a tensor or list of tensors')

        loss_flow = log_vars['loss_flow'].detach()

        if log_vars.get('loss_occ') is not None:
            loss_occ = log_vars['loss_occ'].detach()

            loss_flow_weight = 1.
            loss_occ_weight = 1.

            if loss_flow.data > loss_occ.data and loss_occ.data != 0.:
                loss_flow_weight = 1.
                loss_occ_weight = loss_flow / loss_occ

            elif loss_flow.data < loss_occ.data and loss_flow.data != 0.:
                loss_flow_weight = loss_occ / loss_flow
                loss_occ_weight = 1.

            loss = log_vars['loss_flow'] * loss_flow_weight + log_vars[
                'loss_occ'] * loss_occ_weight
        else:
            loss = log_vars['loss_flow']

        log_vars['loss'] = loss

        for loss_name, loss_value in log_vars.items():
            # reduce loss when distributed training
            if dist.is_available() and dist.is_initialized():
                loss_value = loss_value.data.clone()
                dist.all_reduce(loss_value.div_(dist.get_world_size()))
            log_vars[loss_name] = loss_value.item()

        return loss, log_vars



# Copyright (c) OpenMMLab. All rights reserved.
from typing import Tuple

from mmengine.config import Config
from torch import Tensor

from mmflow.registry import MODELS
from mmflow.utils import OptSampleList, SampleList, TensorDict
from ..builder import build_encoder
from .pwcnet import PWCNet


@MODELS.register_module()
class FlowNetS(PWCNet):
    """FlowNetS flow estimator."""

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

    def extract_feat(self, imgs: Tensor) -> TensorDict:
        """Extract features from images.

        Args:
            imgs (Tensor): The concatenated input images.

        Returns:
            TensorDict: The feature pyramid extracted from the concatenated
                input images.
        """
        return self.encoder(imgs)

    def loss(self, inputs: Tensor, data_samples: SampleList) -> dict:
        """Forward function for FlowNetS when model training.

        Args:
            inputs (Tensor): Input images of shape (N, 6, H, W).
                img1 is inputs[N, :3, H, W] and img2 is
                inputs[N, 3:, H, W]. These should usually be mean
                centered and std scaled.
            data_samples (list[:obj:`FlowDataSample`]): Each item contains the
                meta information of each image and corresponding annotations.

        Returns:
            TensorDict: The losses of output.
        """

        return self.decoder.loss(self.extract_feat(inputs), data_samples)

    def predict(self, inputs: Tensor, data_samples: SampleList) -> SampleList:
        """Forward function for FlowNetS when model testing.

        Args:
            inputs (Tensor): Input images of shape (N, 6, H, W).
                img1 is inputs[N, :3, H, W] and img2 is
                inputs[N, 3:, H, W]. These should usually be mean
                centered and std scaled.
            data_samples (list[:obj:`FlowDataSample`], optional): Each item
                contains the meta information of each image and corresponding
                annotations. Defaults to None.

        Returns:
            Sequence[FlowDataSample]: The batch of predicted optical flow
                with the same size of images before augmentation.
        """
        return self.decoder.predict(self.extract_feat(inputs), data_samples)

    def _forward(self,
                 inputs: Tensor,
                 data_samples: OptSampleList = None) -> TensorDict:
        """Network forward process. Usually includes backbone, neck and head
        forward without any post-processing.

        Args:
            inputs (Tensor): Input images of shape (N, 6, H, W).
                img1 is batch_inputs[N, :3, H, W] and img2 is
                batch_inputs[N, 3:, H, W]. These should usually be mean
                centered and std scaled.
            data_samples (list[:obj:`FlowDataSample`], optional): Each item
                contains the meta information of each image and corresponding
                annotations. Defaults to None.
        Returns:
            Dict[str, :obj:`FlowDataSample`]: The predicted optical flow
            from level6 to level2.
        """
        return self.decoder(self.extract_feat(inputs))


@MODELS.register_module()
class FlowNetC(PWCNet):
    """FlowNetC flow estimator.

    Args:
        corr_level (str): The level to calculate the correlation.
        corr_encoder (Config): The config of correaltion encoder.
    """

    def __init__(self, corr_level: str, corr_encoder: Config, *args,
                 **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self.corr_level = corr_level
        self.corr_encoder = build_encoder(corr_encoder)

    def extract_feat(self, imgs: Tensor) -> Tuple[TensorDict, TensorDict]:
        """Extract features from images.

        Args:
            imgs (Tensor): The concatenated input images.

        Returns:
            Tuple[TensorDict, TensorDict]: The feature pyramid from the first
                image and the feature pyramid from feature correlation.
        """

        in_channels = self.encoder.in_channels
        img1 = imgs[:, :in_channels, ...]
        img2 = imgs[:, in_channels:, ...]
        feat1 = self.encoder(img1)
        feat2 = self.encoder(img2)
        return feat1, self.corr_encoder(feat1[self.corr_level],
                                        feat2[self.corr_level])



# Copyright (c) OpenMMLab. All rights reserved.
import os
import os.path as osp
from typing import Callable, List, Optional, Sequence, Union

from mmengine.dataset import BaseDataset as MME_BaseDataset

from mmflow.registry import DATASETS
from .utils import get_data_filename, load_ann_info, load_img_info


@DATASETS.register_module()
class Sintel(MME_BaseDataset):
    """Sintel optical flow dataset.

    Args:
        pass_style (str): Pass style for Sintel dataset, and it has 2 options
            ['clean', 'final']. Default: 'clean'.
        scene (str, list, optional): Scene in Sintel dataset, if scene is None,
            it means collecting data in all of scene of Sintel dataset.
            Default: None.
        ann_file (str): Annotation file path. Defaults to ''.
        metainfo (dict, optional): Meta information for dataset, such as class
            information. Defaults to None.
        data_root (str, optional): The root directory for ``data_prefix`` and
            ``ann_file``. Defaults to None.
        data_prefix (dict, optional): Prefix for training data. Defaults to
            dict(img=None, ann=None).
        filter_cfg (dict, optional): Config for filter data. Defaults to None.
        indices (int or Sequence[int], optional): Support using first few
            data in annotation file to facilitate training/testing on a smaller
            dataset. Defaults to None which means using all ``data_list``.
        serialize_data (bool, optional): Whether to hold memory using
            serialized objects, when enabled, data loader workers can use
            shared RAM from master process instead of making a copy. Defaults
            to True.
        pipeline (list, optional): Processing pipeline. Defaults to [].
        test_mode (bool, optional): ``test_mode=True`` means in test phase.
            Defaults to False.
        lazy_init (bool, optional): Whether to load annotation during
            instantiation. In some cases, such as visualization, only the meta
            information of the dataset is needed, which is not necessary to
            load annotation file. ``Basedataset`` can skip load annotations to
            save time by set ``lazy_init=False``. Defaults to False.
        max_refetch (int, optional): If ``Basedataset.prepare_data`` get a
            None img. The maximum extra number of cycles to get a valid
            image. Defaults to 1000.
    """
    METAINFO = dict()

    def __init__(self,
                 pass_style: str = 'clean',
                 scene: Optional[Union[str, Sequence[str]]] = None,
                 ann_file: str = '',
                 metainfo: Optional[dict] = None,
                 data_root: Optional[str] = None,
                 data_prefix: dict = dict(),
                 filter_cfg: Optional[dict] = None,
                 indices: Optional[Union[int, Sequence[int]]] = None,
                 serialize_data: bool = True,
                 pipeline: List[Union[dict, Callable]] = [],
                 test_mode: bool = False,
                 lazy_init: bool = False,
                 max_refetch: int = 1000) -> None:

        assert pass_style in (
            'clean',
            'final'), '`pass_style` expected to be \'clean\' or \'final\' '
        f'but got {pass_style}'
        self.pass_style = pass_style

        self.scene = scene

        super().__init__(ann_file, metainfo, data_root, data_prefix,
                         filter_cfg, indices, serialize_data, pipeline,
                         test_mode, lazy_init, max_refetch)

    def load_data_list(self) -> List[dict]:
        """Load ``data_list``

        ``data_list`` can be load from an annotation file named as
        ``self.ann_file`` or by parsing dataset path.

        Returns:
            list[dict]: A list of annotation.
        """  # noqa: E501
        if self.ann_file.endswith('json'):
            # load data_list with annotation file
            return super().load_data_list()
        else:
            # load data_list by path parsing
            self.load_data_info()
            return self.data_list

    def filter_data(self) -> List[dict]:
        """Filter data_list according to ``scene`` and ``pass_style``

        Returns:
            list[int]: Filtered results.
        """
        if self.scene is None:
            new_data_list = [
                data_info for data_info in self.data_list
                if data_info['pass_style'] == self.pass_style
            ]

        else:

            new_data_list = [
                data_info for data_info in self.data_list
                if data_info['scene'] == self.scene
                and data_info['pass_style'] == self.pass_style
            ]
        return new_data_list

    def load_data_info(self) -> None:
        """Load data information, including file path of image1, image2 and
        optical flow."""

        self._get_data_dir()

        img1_filenames = []
        img2_filenames = []
        flow_filenames = []
        occ_filenames = []
        invalid_filenames = []

        def get_filenames(data_dir, data_suffix, img_idx=None):
            data_filenames = []
            for data_dir_ in data_dir:
                data_filenames_ = get_data_filename(data_dir_, data_suffix)
                data_filenames_.sort()
                if img_idx == 1:
                    data_filenames += data_filenames_[:-1]
                elif img_idx == 2:
                    data_filenames += data_filenames_[1:]
                else:
                    data_filenames += data_filenames_
            return data_filenames

        img1_filenames = get_filenames(self.img1_dir, self.img1_suffix, 1)
        img2_filenames = get_filenames(self.img2_dir, self.img2_suffix, 2)
        load_img_info(self.data_list, img1_filenames, img2_filenames)

        if not self.test_mode:
            flow_filenames = get_filenames(self.flow_dir, self.flow_suffix)
            occ_filenames = get_filenames(self.occ_dir, self.occ_suffix)
            invalid_filenames = get_filenames(self.invalid_dir,
                                              self.invalid_suffix, 1)
            load_ann_info(self.data_list, flow_filenames, 'flow_fw_path')
            load_ann_info(self.data_list, occ_filenames, 'occ_fw_path')
            load_ann_info(self.data_list, invalid_filenames, 'invalid_path')

        for i in range(len(self.data_list)):
            self.data_list[i]['pass_style'] = self.pass_style
            self.data_list[i]['scene'] = self.data_list[i]['img1_path'].split(
                os.sep)[-2]
            if self.test_mode:
                self.data_list[i]['flow_fw_path'] = None
                self.data_list[i]['invalid_path'] = None
                self.data_list[i]['occ_fw_path'] = None

    def _get_data_dir(self) -> None:
        """Get the paths for images and optical flow."""
        self.img1_suffix = '.png'
        self.img2_suffix = '.png'
        self.flow_suffix = '.flo'
        self.occ_suffix = '.png'
        self.invalid_suffix = '.png'

        self.subset_dir = osp.join('testing',
                                   'test') if self.test_mode else 'training'

        self.data_root = osp.join(self.data_root, self.subset_dir)

        img_root = osp.join(self.data_root, self.pass_style)
        all_scene = os.listdir(img_root)
        scene = all_scene if self.scene is None else self.scene
        scene = scene if isinstance(scene, (list, tuple)) else [scene]
        assert set(scene).issubset(set(all_scene))
        self.img1_dir = [osp.join(img_root, s) for s in scene]
        self.img2_dir = [osp.join(img_root, s) for s in scene]
        if not self.test_mode:
            flow_root = osp.join(self.data_root, 'flow')
            occ_root = osp.join(self.data_root, 'occlusions')
            invalid_root = osp.join(self.data_root, 'invalid')

            self.flow_dir = [osp.join(flow_root, s) for s in scene]
            self.occ_dir = [osp.join(occ_root, s) for s in scene]
            self.invalid_dir = [osp.join(invalid_root, s) for s in scene]



# Copyright (c) OpenMMLab. All rights reserved.
import argparse
import os.path as osp
from typing import Sequence

import cv2
import mmcv
from mmengine.utils import scandir
from numpy import ndarray

try:
    import imageio
except ImportError:
    imageio = None


def parse_args():
    parser = argparse.ArgumentParser(description='Create GIF for demo')
    parser.add_argument(
        '--frame_dir', type=str, default=None, help='directory of frames')
    parser.add_argument(
        '--resize_factor',
        type=float,
        default=0.5,
        help='resize factor for gif')
    parser.add_argument(
        '--out',
        type=str,
        default='result.gif',
        help='gif path where will be saved')
    parser.add_argument('--duration', default=0.1, help='display interval (s)')

    args = parser.parse_args()
    return args


def create_gif(frames: Sequence[ndarray],
               gif_name: str,
               duration: float = 0.1) -> None:
    """Create gif through imageio.

    Args:
        frames (list[ndarray]): Image frames
        gif_name (str): Saved gif name
        duration (int): Display interval (s). Default: 0.1.
    """
    frames_rgb = []
    for frame in frames:
        frame_rgb = cv2.cvtColor(frame, cv2.COLOR_BGR2RGB)
        frames_rgb.append(frame_rgb)
    if imageio is None:
        raise RuntimeError('imageio is not installed,'
                           'Please use “pip install imageio” to install')
    imageio.mimsave(gif_name, frames_rgb, 'GIF', duration=duration)


def create_frame(frame_dir: str,
                 resize_factor: float = 0.5) -> Sequence[ndarray]:
    """Create gif frame image through matplotlib.

    Args:
        frame_dir (str): The directory of frames.
        resize_factor (float): The factor for rescaling original images.
            Defaults to 0.5.

    Returns:
        list[ndarray]: List of frames.
    """
    frame_files = list(scandir(frame_dir))
    frame_files.sort()
    frame_list = []
    for frame_file in frame_files:
        frame = mmcv.imread(osp.join(frame_dir, frame_file))
        frame_scaled = frame if resize_factor == 1. else mmcv.imrescale(
            frame, scale=resize_factor)
        frame_list.append(frame_scaled)

    return frame_list


def create_video(frames: Sequence[ndarray],
                 out_file: str,
                 duration: float = 0.1) -> None:
    """Create a video to save the optical flow.

    Args:
        frames (list, tuple): Image frames.
        out_file (str): The output file to save visualized flow map.
        duration (int): Display interval (s). Default: 0.1.
    """
    fourcc = cv2.VideoWriter_fourcc(*'mp4v')
    fps = 1 / duration
    H, W = frames[0].shape[:2]
    video_writer = cv2.VideoWriter(out_file, fourcc, fps, (W, H), True)

    for frame in frames:
        video_writer.write(frame)
    video_writer.release()


def main():
    args = parse_args()
    frames = create_frame(args.frame_dir, args.resize_factor)
    if args.out[-3:] == 'gif':
        create_gif(frames, args.out, args.duration)
    else:
        create_video(frames, args.out, args.duration)


if __name__ == '__main__':
    main()



# Copyright (c) OpenMMLab. All rights reserved.
from .flow_io import (flow_from_bytes, read_flow, read_flow_kitti, read_pfm,
                      render_color_wheel, visualize_flow, write_flow,
                      write_flow_kitti)
from .image import adjust_gamma, adjust_hue
from .utils import get_data_filename, load_ann_info, load_img_info

__all__ = [
    'read_flow', 'write_flow', 'visualize_flow', 'render_color_wheel',
    'read_flow_kitti', 'write_flow_kitti', 'adjust_hue', 'adjust_gamma',
    'read_pfm', 'flow_from_bytes', 'get_data_filename', 'load_ann_info',
    'load_img_info'
]



model = dict(
    type='FlowNetC',
    data_preprocessor=dict(
        type='FlowDataPreprocessor',
        mean=[0., 0., 0.],
        std=[255., 255., 255.],
        bgr_to_rgb=False,
        sigma_range=(0, 0.04),
        clamp_range=(0., 1.)),
    encoder=dict(
        type='FlowNetEncoder',
        in_channels=3,
        pyramid_levels=['level1', 'level2', 'level3'],
        out_channels=(64, 128, 256),
        kernel_size=(7, 5, 5),
        strides=(2, 2, 2),
        num_convs=(1, 1, 1),
        dilations=(1, 1, 1),
        act_cfg=dict(type='LeakyReLU', negative_slope=0.1),
    ),
    corr_level='level3',
    corr_encoder=dict(
        type='CorrEncoder',
        in_channels=473,
        pyramid_levels=['level3', 'level4', 'level5', 'level6'],
        kernel_size=(3, 3, 3, 3),
        num_convs=(1, 2, 2, 2),
        out_channels=(256, 512, 512, 1024),
        redir_in_channels=256,
        redir_channels=32,
        strides=(1, 2, 2, 2),
        dilations=(1, 1, 1, 1),
        corr_cfg=dict(
            type='Correlation',
            kernel_size=1,
            max_displacement=10,
            stride=1,
            padding=0,
            dilation_patch=2),
        scaled=False,
        conv_cfg=None,
        norm_cfg=None,
        act_cfg=dict(type='LeakyReLU', negative_slope=0.1),
    ),
    decoder=dict(
        type='FlowNetCDecoder',
        in_channels=dict(
            level6=1024, level5=1026, level4=770, level3=386, level2=194),
        out_channels=dict(level6=512, level5=256, level4=128, level3=64),
        deconv_bias=True,
        pred_bias=True,
        upsample_bias=True,
        norm_cfg=None,
        flow_loss=dict(
            type='MultiLevelEPE',
            p=2,
            reduction='sum',
            weights={
                'level2': 0.005,
                'level3': 0.01,
                'level4': 0.02,
                'level5': 0.08,
                'level6': 0.32
            }),
        act_cfg=dict(type='LeakyReLU', negative_slope=0.1)),
    init_cfg=[
        dict(
            type='Kaiming',
            layer=['Conv2d', 'ConvTranspose2d'],
            a=0.1,
            mode='fan_in',
            nonlinearity='leaky_relu',
            bias=0),
        dict(type='Constant', layer='BatchNorm2d', val=1, bias=0)
    ],  # model training and testing settings
    train_cfg=dict(),
    test_cfg=dict())
randomness = dict(seed=0, diff_rank_seed=True)



# Copyright (c) OpenMMLab. All rights reserved.
from typing import Optional

import torch
import torch.nn.functional as F
from torch import Tensor


def weighted_ssim(x: Tensor,
                  y: Tensor,
                  weight: Optional[Tensor] = None,
                  c1=0.01**2,
                  c2=0.03**2,
                  weight_epsilon=0.01):
    """Computes a weighted structured image similarity.

    This function is modified from
    https://github.com/google-research/google-research/blob/master/uflow/uflow_utils.py
    Copyright 2022 The Google Research Authors.

    Args:
        x (Tensor): A Tensor representing a batch of images, of shape
            [B, C, H, W].
        y (Tensor): A Tensor representing a batch of images, of shape
            [B, C, H, W].
        weight (Tensor, optional): A Tensor of shape [H, W], representing
            the weight of each pixel in both images when we come to calculate
            moments (means and correlations). Defaults to None.
        c1 (float): A floating point number, regularizes division by zero of
            the means. Defaults to 0.01**2.
        c2 (float): A floating point number, regularizes division by zero of
            the means. Defaults to 0.03 ** 2.
        weight_epsilon (float): A floating point number, used to regularize
            division by the weight. Defaults to 0.01.
    Returns:
        A tuple of two Tensors. First, of shape [B, C, H-2, W-2], is scalar
        similarity loss oer pixel per channel. It is needed so that we know
        how much to weigh each pixel in the first tensor. For example, if
        ``weight`` was very small in some area of the images, the first tensor
        will still assign a loss to these pixels, but we shouldn't take the
        result too seriously.
    """
    if c1 == float('inf') and c2 == float('inf'):
        raise ValueError(
            'Both c1 and c2 are infinite, SSIM loss is zero. This is '
            'likely unintended.')
    _, _, H, W = x.shape

    if weight is None:
        weight = torch.ones((H, W)).to(x)
    else:
        assert weight.shape == (H, W), \
                f'image shape is {(H, W)}, but weight shape is {weight.shape}'
    weight = weight[None, None, ...]
    average_pooled_weight = F.avg_pool2d(weight, (3, 3), stride=(1, 1))
    weight_plus_epsilon = weight + weight_epsilon
    inverse_average_pooled_weight = 1.0 / (
        average_pooled_weight + weight_epsilon)

    def weighted_avg_pool3x3(z):
        weighted_avg = F.avg_pool2d(
            z * weight_plus_epsilon, (3, 3), stride=(1, 1))
        return weighted_avg * inverse_average_pooled_weight

    mu_x = weighted_avg_pool3x3(x)
    mu_y = weighted_avg_pool3x3(y)
    sigma_x = weighted_avg_pool3x3(x**2) - mu_x**2
    sigma_y = weighted_avg_pool3x3(y**2) - mu_y**2
    sigma_xy = weighted_avg_pool3x3(x * y) - mu_x * mu_y
    if c1 == float('inf'):
        ssim_n = (2 * sigma_xy + c2)
        ssim_d = (sigma_x + sigma_y + c2)
    elif c2 == float('inf'):
        ssim_n = 2 * mu_x * mu_y + c1
        ssim_d = mu_x**2 + mu_y**2 + c1
    else:
        ssim_n = (2 * mu_x * mu_y + c1) * (2 * sigma_xy + c2)
        ssim_d = (mu_x**2 + mu_y**2 + c1) * (sigma_x + sigma_y + c2)
    result = ssim_n / ssim_d
    return torch.clamp((1 - result) / 2, 0, 1)



# Copyright (c) OpenMMLab. All rights reserved.
from .context_net import ContextNet
from .flownet_decoder import FlowNetCDecoder, FlowNetSDecoder
from .gma_decoder import GMADecoder
from .irr_refine import FlowRefine, OccRefine, OccShuffleUpsample
from .irrpwc_decoder import IRRPWCDecoder
from .liteflownet_decoder import NetE
from .maskflownet_decoder import MaskFlowNetDecoder, MaskFlowNetSDecoder
from .pwcnet_decoder import PWCNetDecoder
from .raft_decoder import RAFTDecoder

__all__ = [
    'FlowNetCDecoder', 'FlowNetSDecoder', 'PWCNetDecoder',
    'MaskFlowNetSDecoder', 'NetE', 'ContextNet', 'RAFTDecoder', 'FlowRefine',
    'OccRefine', 'OccShuffleUpsample', 'IRRPWCDecoder', 'MaskFlowNetDecoder',
    'GMADecoder'
]



# Copyright (c) OpenMMLab. All rights reserved.
import argparse
import json
import os.path as osp
from glob import glob

from mmengine.utils import mkdir_or_exist


def parse_args():
    parser = argparse.ArgumentParser(description='HD1K dataset preparation')
    parser.add_argument(
        '--data-root',
        type=str,
        default='data/hd1k',
        help='Directory for dataset.')

    parser.add_argument(
        '--save-dir',
        type=str,
        default='data/hd1k/',
        help='Directory to save the annotation files for HD1K dataset')
    args = parser.parse_args()

    return args


def main():
    args = parse_args()

    img1_dir = osp.join(args.data_root, 'hd1k_input/image_2')
    flow_dir = osp.join(args.data_root, 'hd1k_flow_gt/flow_occ')

    img1_filenames = []
    img2_filenames = []
    flow_filenames = []

    seq_ix = 0
    while True:
        flows = sorted(glob(osp.join(flow_dir, '%06d_*.png' % seq_ix)))
        images = sorted(glob(osp.join(img1_dir, '%06d_*.png' % seq_ix)))

        if len(flows) == 0:
            break

        for i in range(len(flows) - 1):
            flow_filenames += [flows[i]]
            img1_filenames += [images[i]]
            img2_filenames += [images[i + 1]]

        seq_ix += 1

    train_list = []

    for i in range(len(flow_filenames)):
        data_info = dict(
            img1_path=osp.join(img1_filenames[i]),
            img2_path=osp.join(img2_filenames[i]),
            flow_fw_path=osp.join(flow_filenames[i]))

        train_list.append(data_info)
    mkdir_or_exist(args.save_dir)
    with open(osp.join(args.save_dir, 'train.json'), 'w') as jsonfile:
        json.dump({'metainfo': {}, 'data_list': train_list}, jsonfile)


if __name__ == '__main__':
    main()



# Copyright (c) OpenMMLab. All rights reserved.
import os.path as osp
import warnings
from typing import Sequence

from mmengine.hooks import Hook
from mmengine.runner import Runner
from mmengine.visualization import Visualizer

from mmflow.registry import HOOKS
from mmflow.structures import FlowDataSample


@HOOKS.register_module()
class FlowVisualizationHook(Hook):
    """Flow Visualization Hook. Used to visualize validation and testing
    process prediction results.

    In the testing phase:
       If ``show`` is True, it means that only the prediction results are
        visualized without storing data, so ``vis_backends`` needs to
        be excluded.

    Args:
        draw (bool): whether to draw prediction results. If it is False,
            it means that no drawing will be done. Defaults to False.
        interval (int): The interval of visualization. Defaults to 50.
        show (bool): Whether to display the drawn image. Default to False.
        wait_time (int): Delay in milliseconds. 0 is the special
            value that means "forever". Defaults to 0.
    """

    def __init__(self,
                 draw: bool = False,
                 interval: int = 50,
                 show: bool = False,
                 wait_time: int = 0):
        self._visualizer: Visualizer = Visualizer.get_current_instance()
        self.interval = interval
        self.show = show
        if self.show:
            # No need to think about vis backends.
            self._visualizer._vis_backends = {}
            warnings.warn('The show is True, it means that only '
                          'the prediction results are visualized '
                          'without storing data, so vis_backends '
                          'needs to be excluded.')

        self.wait_time = wait_time
        self.draw = draw
        if not self.draw:
            warnings.warn('The draw is False, it means that the '
                          'hook for visualization will not take '
                          'effect. The results will NOT be '
                          'visualized or stored.')

    def _after_iter(self,
                    runner: Runner,
                    batch_idx: int,
                    data_batch: Sequence[dict],
                    outputs: Sequence[FlowDataSample],
                    mode: str = 'val') -> None:
        """Run after every ``self.interval`` validation iterations.

        Args:
            runner (:obj:`Runner`): The runner of the validation process.
            batch_idx (int): The index of the current batch in the val loop.
            data_batch (Sequence[dict]): Data from dataloader.
            outputs (Sequence[:obj:`FlowDataSample`]): Outputs from model.
            mode (str): mode (str): Current mode of runner. Defaults to 'val'.
        """
        if self.draw is False or mode == 'train':
            return

        if self.every_n_inner_iters(batch_idx, self.interval):
            for output in outputs:
                img1_path = output.metainfo['img1_path']
                img2_path = output.metainfo['img2_path']
                window_name = f'{mode}_{osp.basename(img1_path)}' \
                              f'_{osp.basename(img2_path)}'

                self._visualizer.add_datasample(
                    window_name,
                    image=None,
                    data_sample=output,
                    show=self.show,
                    wait_time=self.wait_time,
                    step=runner.iter)



# Copyright (c) OpenMMLab. All rights reserved.

import os.path as osp
from typing import Optional, Sequence, Union

from mmengine.utils import scandir


def get_data_filename(
        data_dirs: Union[Sequence[str], str],
        suffix: Optional[str] = None,
        exclude: Optional[Sequence[str]] = None) -> Sequence[str]:
    """Get file name from data directory.

    Args:
        data_dirs (list, str): the directory of data
        suffix (str, optional): the suffix for data file. Defaults to None.
        exclude (list, optional): list of files will be excluded.
    Returns:
        list: the list of data file.
    """

    if data_dirs is None:
        return None
    data_dirs = data_dirs \
        if isinstance(data_dirs, (list, tuple)) else [data_dirs]

    suffix = '' if suffix is None else suffix
    if exclude is None:
        exclude = []
    else:
        assert isinstance(exclude, (list, tuple))

    files = []
    for data_dir in data_dirs:
        for f in scandir(data_dir, suffix=suffix):
            if f not in exclude:
                files.append(osp.join(data_dir, f))
    files.sort()
    return files



# Copyright (c) OpenMMLab. All rights reserved.
from .builder import build_dataset
from .chairssdhom import ChairsSDHom
from .flyingchairs import FlyingChairs
from .flyingchairsocc import FlyingChairsOcc
from .flyingthings3d import FlyingThings3D
from .flyingthings3d_subset import FlyingThings3DSubset
from .hd1k import HD1K
from .kiiti2012 import KITTI2012
from .kitti2015 import KITTI2015
from .samplers import DistributedSampler, MixedBatchDistributedSampler
from .sintel import Sintel
from .transforms import (ColorJitter, Compose, Erase, GaussianNoise, InputPad,
                         InputResize, Normalize, PackFlowInputs,
                         PhotoMetricDistortion, RandomAffine, RandomCrop,
                         RandomFlip, Rerange, SpacialTransform, Validation)
from .utils import (read_flow, read_flow_kitti, render_color_wheel,
                    visualize_flow, write_flow, write_flow_kitti)

__all__ = [
    'build_dataset', 'FlyingChairs', 'SpacialTransform', 'Validation', 'Erase',
    'RandomFlip', 'Normalize', 'Rerange', 'RandomCrop', 'ColorJitter',
    'PhotoMetricDistortion', 'RandomAffine', 'MixedBatchDistributedSampler',
    'DistributedSampler', 'read_flow', 'visualize_flow', 'write_flow',
    'InputResize', 'write_flow_kitti', 'read_flow_kitti', 'GaussianNoise',
    'Compose', 'InputPad', 'FlyingThings3DSubset', 'FlyingThings3D', 'Sintel',
    'KITTI2012', 'KITTI2015', 'render_color_wheel', 'PackFlowInputs',
    'FlyingChairsOcc', 'HD1K', 'ChairsSDHom'
]



# Copyright (c) OpenMMLab. All rights reserved.
from typing import Optional, Sequence, Union

from mmflow.registry import MODELS
from .base_encoder import BasicEncoder


@MODELS.register_module()
class PWCNetEncoder(BasicEncoder):
    """The feature extraction sub-module in PWC-Net.

    Args:
        in_channels (int): Number of input channels.
        pyramid_levels (Sequence[str]): The list of feature pyramid that are
            the keys for output dict.
        net_type (str): The type of this sub-module, if net_type is Basic, the
            the number of convolution layers of each level is 3, if net_type is
            Small, the the number of convolution layers of each level is 2.
        out_channels (Sequence[int]): List of numbers of output
            channels of each pyramid level.
            Default: (16, 32, 64, 96, 128, 196).
        strides (Sequence[int]): List of strides of each pyramid level.
            Default: (2, 2, 2, 2, 2, 2).
        dilations (Sequence[int]): List of dilation of each pyramid level.
            Default: (1, 1, 1, 1, 1, 1).
        conv_cfg (dict, optional): Config dict for convolution layer.
            Default: None.
        norm_cfg (dict, optional): Config dict for each normalization layer.
            Default: None.
        act_cfg (dict): Config dict for each activation layer in ConvModule.
            Default: dict(type='LeakyReLU', negative_slope=0.1).
        init_cfg (dict, optional): Config of weights initialization. Default:
            None.
    """
    _arch_settings = {'Basic': (3, 3, 3, 3, 3, 3), 'Small': (2, 2, 2, 2, 2, 2)}

    def __init__(self,
                 in_channels: int,
                 pyramid_levels: Sequence[str],
                 net_type: str = 'Basic',
                 out_channels: Sequence[int] = (16, 32, 64, 96, 128, 196),
                 strides: Sequence[int] = (2, 2, 2, 2, 2, 2),
                 dilations: Sequence[int] = (1, 1, 1, 1, 1, 1),
                 conv_cfg: Optional[dict] = None,
                 norm_cfg: Optional[dict] = None,
                 act_cfg: dict = dict(type='LeakyReLU', negative_slope=0.1),
                 init_cfg: Optional[Union[list, dict]] = None) -> None:

        if net_type not in self._arch_settings:
            raise KeyError(f'invalid net type {net_type} for PWC-Net')

        num_convs = self._arch_settings[net_type]

        super().__init__(
            in_channels=in_channels,
            pyramid_levels=pyramid_levels,
            num_convs=num_convs,
            out_channels=out_channels,
            strides=strides,
            dilations=dilations,
            conv_cfg=conv_cfg,
            norm_cfg=norm_cfg,
            act_cfg=act_cfg,
            init_cfg=init_cfg)



# Copyright (c) OpenMMLab. All rights reserved.
from typing import Dict, List, Optional, Tuple, Union

import mmcv
import numpy as np
from mmcv import sparse_flow_from_bytes
from mmcv.transforms import BaseTransform
from mmengine.fileio import FileClient

from mmflow.registry import TRANSFORMS
from ..utils import flow_from_bytes


@TRANSFORMS.register_module()
class LoadImageFromFile(BaseTransform):
    """Load image1 and image2 from file.

    Required Keys:

    - img1_path
    - img2_path

    Modified Keys:

    - img1
    - img2
    - img_shape
    - ori_shape

    Args:
        to_float32 (bool): Whether to convert the loaded image to a float32
            numpy array. If set to False, the loaded image is an uint8 array.
            Defaults to False.
        color_type (str): The flag argument for :func:``mmcv.imfrombytes``.
            Defaults to 'color'.
        imdecode_backend (str): The image decoding backend type. The backend
            argument for :func:``mmcv.imfrombytes``.
            See :func:``mmcv.imfrombytes`` for details.
            Defaults to 'cv2'.
        file_client_args (dict): Arguments to instantiate a FileClient.
            See :class:`mmengine.fileio.FileClient` for details.
            Defaults to ``dict(backend='disk')``.
    """

    def __init__(self,
                 to_float32: bool = False,
                 color_type: str = 'color',
                 file_client_args: dict = dict(backend='disk'),
                 imdecode_backend: str = 'cv2') -> None:
        super().__init__()
        self.to_float32 = to_float32
        self.color_type = color_type
        self.file_client_args = file_client_args.copy()
        self.file_client = FileClient(**self.file_client_args)
        self.imdecode_backend = imdecode_backend

    def transform(self,
                  results: Dict) -> Optional[Union[Dict, Tuple[List, List]]]:
        """Call function to load image and get image meta information.

        Args:
            results (dict): Result dict from :obj:`mmflow.BaseDataset`.

        Returns:
            dict: The dict contains loaded image and meta information.
        """
        filename1 = results['img1_path']
        img1_bytes = self.file_client.get(filename1)
        img1 = mmcv.imfrombytes(
            img1_bytes, flag=self.color_type, backend=self.imdecode_backend)

        filename2 = results['img2_path']
        img2_bytes = self.file_client.get(filename2)
        img2 = mmcv.imfrombytes(
            img2_bytes, flag=self.color_type, backend=self.imdecode_backend)

        if self.to_float32:
            img1 = img1.astype(np.float32)
            img2 = img2.astype(np.float32)

        results['img1'] = img1
        results['img2'] = img2
        results['img_shape'] = img1.shape[:2]
        results['ori_shape'] = img1.shape[:2]

        return results

    def __repr__(self) -> str:
        repr_str = self.__class__.__name__
        repr_str += f'(to_float32={self.to_float32},'
        repr_str += f"color_type='{self.color_type}',"
        repr_str += f"imdecode_backend='{self.imdecode_backend}')"
        return repr_str


@TRANSFORMS.register_module()
class LoadAnnotations(BaseTransform):
    """Load optical flow from file.

    The annotation format is as the following:

    .. code-block:: python

        {
            # Filename of optical flow ground truth file.
            'flow_fw_path': 'a/b/c',
            'flow_bw_path': 'a/b/c',
            'occ_bw_path': 'a/b/c',
            'occ_fw_path': 'a/b/c',

        }

    After this module, the annotation has been changed to the format below:

    .. code-block:: python

        {

            'gt_flow_fw': np.ndarray (H, W, 2)
            'gt_flow_bw': np.ndarray (H, W, 2)
            'gt_occ_fw': np.ndarray (H, W)
            'gt_occ_bw': np.ndarray (H, W)
            'gt_valid_fw': np.ndarray (H, W)
            'gt_valid_bw': np.ndarray (H, W)

        }

    Required Keys:

    - flow_fw_path
    - flow_bw_path (optional)
    - occ_fw_path (optional)
    - occ_bw_path (optional)

    Added Keys:

    - gt_flow_fw (np.float32)
    - gt_flow_bw (np.float32, optional)
    - gt_occ_fw (np.uint, optional)
    - gt_occ_bw (np.uint, optional)
    - gt_valid_fw (np.float32, optional)
    - gt_valid_bw (np.float32, optional)

    Args:
        with_occ (bool): whether to parse and load occlusion mask.
            Default to False.
        sparse (bool): whether the flow is sparse. Default to False.
        file_client_args (dict): Arguments to instantiate a FileClient.
            See :class:`mmengine.fileio.FileClient` for details.
            Defaults to ``dict(backend='disk')``.
    """

    def __init__(
            self,
            with_occ: bool = False,
            sparse: bool = False,
            file_client_args: dict = dict(backend='disk'),
    ) -> None:

        self.with_occ = with_occ
        self.sparse = sparse
        self.file_client_args = file_client_args
        self.file_client = None

    def transform(self, results: Dict) -> Dict:
        """Call function to load optical flow and occlusion mask (optional).

        Args:
            results (dict): Result dict from :obj:`mmflow.Dataset`.

        Returns:
            dict: The dict contains loaded annotation data.
        """

        if self.file_client is None:
            self.file_client = FileClient(**self.file_client_args)

        if self.sparse:
            results = self._load_sparse_flow(results)
        else:
            results = self._load_flow(results)
        if self.with_occ:
            results = self._load_occ(results)

        results['sparse'] = self.sparse

        return results

    def _load_flow(self, results: dict) -> dict:
        """load dense optical flow function.

        Args:
            results (dict): Result dict from :obj:`mmflow.BaseDataset`.

        Returns:
            dict: The dict contains loaded annotation data.
        """

        flow_fw_filename = results.get('flow_fw_path', None)
        flow_bw_filename = results.get('flow_bw_path', None)

        if flow_fw_filename is not None:
            flow_fw_bytes = self.file_client.get(flow_fw_filename)
            flow_fw = flow_from_bytes(flow_fw_bytes, flow_fw_filename[-3:])
        else:
            flow_fw = None

        if flow_bw_filename is not None:
            flow_bw_bytes = self.file_client.get(flow_bw_filename)
            flow_bw = flow_from_bytes(flow_bw_bytes, flow_bw_filename[-3:])
        else:
            flow_bw = None
        results['gt_flow_fw'] = flow_fw
        results['gt_flow_bw'] = flow_bw

        return results

    def _load_sparse_flow(self, results: dict) -> dict:
        """load sparse optical flow function.

        Args:
            results (dict): Result dict from :obj:`mmflow.BaseDataset`.

        Returns:
            dict: The dict contains loaded annotation data.
        """
        flow_fw_filename = results.get('flow_fw_path', None)

        if flow_fw_filename is not None:
            flow_fw_bytes = self.file_client.get(flow_fw_filename)
            flow_fw, valid_fw = sparse_flow_from_bytes(flow_fw_bytes)
        else:
            flow_fw = None
            valid_fw = None

        results['gt_flow_fw'] = flow_fw
        results['gt_flow_bw'] = None
        # sparse flow dataset don't include backward flow
        results['gt_valid_fw'] = valid_fw
        results['gt_valid_bw'] = None

        return results

    def _load_occ(self, results: dict) -> dict:
        """load annotation function.

        Args:
            results (dict): Result dict from :obj:`mmflow.BaseDataset`.

        Returns:
            dict: The dict contains loaded annotation data.
        """
        occ_fw_filename = results.get('occ_fw_path', None)
        occ_bw_filename = results.get('occ_bw_path', None)

        if occ_fw_filename is not None:
            occ_fw_bytes = self.file_client.get(occ_fw_filename)
            occ_fw = (mmcv.imfrombytes(occ_fw_bytes, flag='grayscale') /
                      255).astype(np.float32)
        else:
            occ_fw = None
        if occ_bw_filename is not None:
            occ_bw_bytes = self.file_client.get(occ_bw_filename)
            occ_bw = (mmcv.imfrombytes(occ_bw_bytes, flag='grayscale') /
                      255).astype(np.float32)
        else:
            occ_bw = None
        results['gt_occ_fw'] = occ_fw
        results['gt_occ_bw'] = occ_bw

        return results

    def __repr__(self) -> str:

        repr_str = self.__class__.__name__
        repr_str += f'(with_occ={self.with_occ},'
        repr_str += f"sparse='{self.sparse}',"
        repr_str += f"file_client_args='{self.file_client_args}')"

        return repr_str


@TRANSFORMS.register_module()
class LoadImageFromWebcam(LoadImageFromFile):
    """Load an image from webcam.

    Similar with :obj:`LoadImageFromFile`, but the image read from webcam is in
    ``results['img']``.
    """

    def __call__(self, results: dict) -> dict:
        """Call function to add image meta information.

        Args:
            results (dict): Result dict with Webcam read image in
                ``results['img']``.
        Returns:
            dict: The dict contains loaded image and meta information.
        """

        img1 = results['img1']
        img2 = results['img2']
        if self.to_float32:
            img1 = img1.astype(np.float32)
            img2 = img2.astype(np.float32)

        results['img1_path'] = None
        results['img2_path'] = None
        results['img1'] = img1
        results['img2'] = img2
        results['img_shape'] = img1.shape[:2]
        results['ori_shape'] = img1.shape[:2]

        return results



model = dict(
    type='FlowNetS',
    data_preprocessor=dict(
        type='FlowDataPreprocessor',
        mean=[0., 0., 0.],
        std=[255., 255., 255.],
        bgr_to_rgb=False,
        sigma_range=(0, 0.04),
        clamp_range=(0., 1.)),
    encoder=dict(
        type='FlowNetSDEncoder',
        in_channels=6,
        plugin_channels=64,
        pyramid_levels=[
            'level1', 'level2', 'level3', 'level4', 'level5', 'level6'
        ],
        num_convs=(2, 2, 2, 2, 2, 2),
        out_channels=((64, 128), 128, 256, 512, 512, 1024),
        kernel_size=3,
        strides=(2, 2, 2, 2, 2, 2),
        dilations=(1, 1, 1, 1, 1, 1),
        act_cfg=dict(type='LeakyReLU', negative_slope=0.1),
    ),
    decoder=dict(
        type='FlowNetSDecoder',
        in_channels=dict(
            level6=1024, level5=1026, level4=770, level3=386, level2=194),
        out_channels=dict(level6=512, level5=256, level4=128, level3=64),
        inter_channels=dict(level5=512, level4=256, level3=128, level2=64),
        deconv_bias=True,
        pred_bias=True,
        upsample_bias=True,
        act_cfg=dict(type='LeakyReLU', negative_slope=0.1),
        flow_loss=dict(
            type='MultiLevelEPE',
            p=2,
            reduction='sum',
            weights={
                'level2': 0.005,
                'level3': 0.01,
                'level4': 0.02,
                'level5': 0.08,
                'level6': 0.32
            }),
    ),
    init_cfg=[
        dict(
            type='Kaiming',
            layer=['Conv2d', 'ConvTranspose2d'],
            a=0.1,
            mode='fan_in',
            nonlinearity='leaky_relu',
            bias=0),
        dict(type='Constant', layer='BatchNorm2d', val=1, bias=0)
    ],
    # model training and testing settings
    train_cfg=dict(),
    test_cfg=dict())
randomness = dict(seed=0, diff_rank_seed=True)



# Copyright (c) OpenMMLab. All rights reserved.
from typing import Any, Sequence

from mmengine.optim.scheduler.lr_scheduler import LRSchedulerMixin
from mmengine.optim.scheduler.momentum_scheduler import MomentumSchedulerMixin
from mmengine.optim.scheduler.param_scheduler import INF, _ParamScheduler
from torch.optim import Optimizer

from mmflow.registry import PARAM_SCHEDULERS


@PARAM_SCHEDULERS.register_module()
class MultiStageParamScheduler(_ParamScheduler):
    """Adjust the parameter value of each parameter group by multistage
    scheduler.

    Args:
        optimizer (Optimizer): Wrapped optimizer.
        param_name (str): Name of the parameter to be adjusted, such as
            ``lr``, ``momentum``.
        milestone_params (Sequence[float]): The base params for multi-stages.
            For optical flow tasks, we usually set param_name to ``lr``,
            then the initial learning rate of every stage will be assigned
            by the element in milestone_params.
        milestone_iters (Sequence[int]): The first iterations in different
            stages.
        gammas (Sequence[float]): The list of decay param ratios.
        steps (Sequence[Sequence[int]]): The steps to decay the params
            in stages.
        by_epoch (bool): Whether the scheduled parameters are updated by
            epochs. Optical flow works often use iter-based schedulers.
            Defaults to False.
        kwargs (any): The arguments of _ParamScheduler.
    """

    def __init__(self,
                 optimizer: Optimizer,
                 param_name: str,
                 milestone_params: Sequence[float],
                 milestone_iters: Sequence[int],
                 gammas: Sequence[float],
                 steps: Sequence[Sequence[int]],
                 by_epoch: bool = False,
                 **kwargs: Any):
        assert len(milestone_params) == len(milestone_iters) == len(
            steps
        ) == len(gammas), (
            'For MultiStageLr, lengths of milestones_lr and steps and gammas',
            f'must be equal, but got {len(milestone_params)}, ',
            f'{len(milestone_iters)}, {len(steps)}, and {len(gammas)}')

        for i in range(len(milestone_iters)):
            assert milestone_iters[i] < steps[i][0], (
                'miltestone iter must be less than corresponding step')

        self.milestone_params = milestone_params
        self.steps = steps
        self.gammas = gammas
        self.milestone_iters = milestone_iters

        super().__init__(
            optimizer=optimizer,
            param_name=param_name,
            by_epoch=by_epoch,
            **kwargs)

    @classmethod
    def build_iter_from_epoch(cls,
                              *args,
                              milestone_iters,
                              steps,
                              begin=0,
                              end=INF,
                              by_epoch=True,
                              epoch_length=None,
                              **kwargs):
        """Build an iter-based instance of this scheduler from an epoch-based
        config."""
        assert by_epoch, 'Only epoch-based kwargs whose `by_epoch=True` can ' \
                         'be converted to iter-based.'
        assert epoch_length is not None and epoch_length > 0, \
            f'`epoch_length` must be a positive integer, ' \
            f'but got {epoch_length}.'
        by_epoch = False
        milestone_iters = [i * epoch_length for i in milestone_iters]
        steps = [[j * epoch_length for j in i] for i in steps]
        begin = begin * epoch_length
        if end != INF:
            end = end * epoch_length
        return cls(
            *args,
            milestone_iters=milestone_iters,
            steps=steps,
            begin=begin,
            end=end,
            by_epoch=by_epoch,
            **kwargs)

    def _get_value(self):
        """Compute value using non-chainable form of the scheduler."""
        if self.last_step < self.milestone_iters[0]:
            return [
                group[self.param_name] for group in self.optimizer.param_groups
            ]

        milestone = -1
        for i, milestone_iter in enumerate(self.milestone_iters[1:]):
            if self.last_step < milestone_iter:
                milestone = i
                break

        exp = len(self.steps[milestone])
        for ii, s in enumerate(self.steps[milestone]):
            if self.last_step < s:
                exp = ii
                break

        param = self.milestone_params[milestone] * (
            self.gammas[milestone]**exp)

        return [param for _ in self.optimizer.param_groups]


@PARAM_SCHEDULERS.register_module()
class MultiStageLR(LRSchedulerMixin, MultiStageParamScheduler):
    """Adjust the learning rate of each parameter group by multistage
    scheduler.

    Args:
        optimizer (Optimizer): Wrapped optimizer.
        milestone_params (Sequence[float]): The base lr for multi-stages.
            The initial learning rate of every stage will be assigned
            by the element in milestone_params.
        milestone_iters (Sequence[int]): The first iterations in different
            stages.
        gammas (Sequence[float]): The list of decay lr ratios.
        steps (Sequence[Sequence[int]]): The steps to decay the lrs in stages.
        by_epoch (bool): Whether the scheduled lrs are updated by
            epochs. Optical flow works often use iter-based schedulers.
            Defaults to False.
    """


@PARAM_SCHEDULERS.register_module()
class MultiStageMomentum(MomentumSchedulerMixin, MultiStageParamScheduler):
    """Adjust the momentum value of each parameter group by multistage
    scheduler.

    Args:
        optimizer (Optimizer): Wrapped optimizer.
        milestone_params (Sequence[float]): The base momentum for multi-stages.
            The initial momentum of every stage will be assigned
            by the element in milestone_params.
        milestone_iters (Sequence[int]): The first iterations in different
            stages.
        gammas (Sequence[float]): The list of decay momentum ratios.
        steps (Sequence[Sequence[int]]): The steps to decay the momentum
            in stages.
        by_epoch (bool): Whether the scheduled momentum are updated by
            epochs. Optical flow works often use iter-based schedulers.
            Defaults to False.
    """
